import concurrent.futures
import hashlib
import itertools
import logging
import re
import shutil
import threading
import warnings
import xml.etree.ElementTree as ET
from collections import OrderedDict, defaultdict, namedtuple
from datetime import date, datetime, timedelta
from pathlib import Path
from typing import Any, Dict
from urllib.parse import quote_plus, urljoin

import geojson
import geomet.wkt
import html2text
import requests
from tqdm.auto import tqdm

from sentinelsat.exceptions import (
    InvalidChecksumError,
    InvalidKeyError,
    LTAError,
    LTATriggered,
    QueryLengthError,
    QuerySyntaxError,
    SentinelAPIError,
    ServerError,
    UnauthorizedError,
)
from . import __version__ as sentinelsat_version


class SentinelAPI:
    """Class to connect to Copernicus Open Access Hub, search and download imagery.

    Parameters
    ----------
    user : string
        username for DataHub
        set to None to use ~/.netrc
    password : string
        password for DataHub
        set to None to use ~/.netrc
    api_url : string, optional
        URL of the DataHub
        defaults to 'https://apihub.copernicus.eu/apihub'
    show_progressbars : bool
        Whether progressbars should be shown or not, e.g. during download. Defaults to True.
    timeout : float or tuple, default 60
        How long to wait for DataHub response (in seconds).
        Tuple (connect, read) allowed.
        Set to None to wait indefinitely.

    Attributes
    ----------
    session : requests.Session
        Session to connect to DataHub
    api_url : str
        URL to the DataHub
    page_size : int
        Number of results per query page.
        Current value: 100 (maximum allowed on ApiHub)
    timeout : float or tuple
        How long to wait for DataHub response (in seconds).
    """

    logger = logging.getLogger("sentinelsat.SentinelAPI")

    def __init__(
        self,
        user,
        password,
        api_url="https://apihub.copernicus.eu/apihub/",
        show_progressbars=True,
        timeout=60,
    ):
        self.session = requests.Session()
        if user and password:
            self.session.auth = (user, password)
        self.api_url = api_url if api_url.endswith("/") else api_url + "/"
        self.page_size = 100
        self.user_agent = "sentinelsat/" + sentinelsat_version
        self.session.headers["User-Agent"] = self.user_agent
        self.session.timeout = timeout
        self.show_progressbars = show_progressbars
        self._dhus_version = None
        # For unit tests
        self._last_query = None
        self._last_response = None

    @staticmethod
    def _api2dhus_url(api_url):
        url = re.sub("apihub/$", "dhus/", api_url)
        url = re.sub("apihub.copernicus.eu", "scihub.copernicus.eu", url)
        return url

    def _req_dhus_stub(self):
        try:
            resp = self.session.get(self.api_url + "api/stub/version")
            resp.raise_for_status()
        except requests.exceptions.HTTPError as err:
            self.logger.error("HTTPError: %s", err)
            self.logger.error("Are you trying to get the DHuS version of APIHub?")
            self.logger.error("Trying again after conversion to DHuS URL")
            resp = self.session.get(self._api2dhus_url(self.api_url) + "api/stub/version")
            resp.raise_for_status()
        return resp.json()["value"]

    @property
    def dhus_version(self):
        if self._dhus_version is None:
            self._dhus_version = self._req_dhus_stub()
        return self._dhus_version

    def query(
        self,
        area=None,
        date=None,
        raw=None,
        area_relation="Intersects",
        order_by=None,
        limit=None,
        offset=0,
        **keywords
    ):
        """Query the OpenSearch API with the coordinates of an area, a date interval
        and any other search keywords accepted by the API.

        Parameters
        ----------
        area : str, optional
            The area of interest formatted as a Well-Known Text string.
        date : tuple of (str or datetime) or str, optional
            A time interval filter based on the Sensing Start Time of the products.
            Expects a tuple of (start, end), e.g. ("NOW-1DAY", "NOW").
            The timestamps can be either a Python datetime or a string in one of the
            following formats:

                - yyyyMMdd
                - yyyy-MM-ddThh:mm:ss.SSSZ (ISO-8601)
                - yyyy-MM-ddThh:mm:ssZ
                - NOW
                - NOW-<n>DAY(S) (or HOUR(S), MONTH(S), etc.)
                - NOW+<n>DAY(S)
                - yyyy-MM-ddThh:mm:ssZ-<n>DAY(S)
                - NOW/DAY (or HOUR, MONTH etc.) - rounds the value to the given unit

            Alternatively, an already fully formatted string such as "[NOW-1DAY TO NOW]" can be
            used as well.
        raw : str, optional
            Additional query text that will be appended to the query.
        area_relation : {'Intersects', 'Contains', 'IsWithin'}, optional
            What relation to use for testing the AOI. Case insensitive.

                - Intersects: true if the AOI and the footprint intersect (default)
                - Contains: true if the AOI is inside the footprint
                - IsWithin: true if the footprint is inside the AOI

        order_by: str, optional
            A comma-separated list of fields to order by (on server side).
            Prefix the field name by '+' or '-' to sort in ascending or descending order,
            respectively. Ascending order is used if prefix is omitted.
            Example: "cloudcoverpercentage, -beginposition".
        limit: int, optional
            Maximum number of products returned. Defaults to no limit.
        offset: int, optional
            The number of results to skip. Defaults to 0.
        **keywords
            Additional keywords can be used to specify other query parameters,
            e.g. `relativeorbitnumber=70`.
            See https://scihub.copernicus.eu/twiki/do/view/SciHubUserGuide/3FullTextSearch
            for a full list.


        Range values can be passed as two-element tuples, e.g. `cloudcoverpercentage=(0, 30)`.
        `None` can be used in range values for one-sided ranges, e.g. `orbitnumber=(16302, None)`.
        Ranges with no bounds (`orbitnumber=(None, None)`) will not be included in the query.

        Multiple values for the same query parameter can be provided as sets and will be handled as
        logical OR, e.g. `orbitnumber={16302, 1206}`.

        The time interval formats accepted by the `date` parameter can also be used with
        any other parameters that expect time intervals (that is: 'beginposition', 'endposition',
        'date', 'creationdate', and 'ingestiondate').

        Returns
        -------
        dict[string, dict]
            Products returned by the query as a dictionary with the product ID as the key and
            the product's attributes (a dictionary) as the value.
        """
        query = self.format_query(area, date, raw, area_relation, **keywords)

        if query.strip() == "":
            # An empty query should return the full set of products on the server, which is a bit unreasonable.
            # The server actually raises an error instead and it's better to fail early in the client.
            raise ValueError("Empty query.")

        # check query length - often caused by complex polygons
        if self.check_query_length(query) > 1.0:
            self.logger.warning(
                "The query string is too long and will likely cause a bad DHuS response."
            )

        self.logger.debug(
            "Running query: order_by=%s, limit=%s, offset=%s, query=%s",
            order_by,
            limit,
            offset,
            query,
        )
        formatted_order_by = _format_order_by(order_by)
        response, count = self._load_query(query, formatted_order_by, limit, offset)
        self.logger.info("Found %s products", count)
        return _parse_opensearch_response(response)

    @staticmethod
    def format_query(area=None, date=None, raw=None, area_relation="Intersects", **keywords):
        """Create a OpenSearch API query string."""
        if area_relation.lower() not in {"intersects", "contains", "iswithin"}:
            raise ValueError("Incorrect AOI relation provided ({})".format(area_relation))

        # Check for duplicate keywords
        kw_lower = {x.lower() for x in keywords}
        if (
            len(kw_lower) != len(keywords)
            or (date is not None and "beginposition" in kw_lower)
            or (area is not None and "footprint" in kw_lower)
        ):
            raise ValueError(
                "Query contains duplicate keywords. Note that query keywords are case-insensitive."
            )

        query_parts = []

        if date is not None:
            keywords["beginPosition"] = date

        for attr, value in sorted(keywords.items()):
            if isinstance(value, set):
                if len(value) == 0:
                    continue
                sub_parts = []
                for sub_value in value:
                    sub_value = _format_query_value(attr, sub_value)
                    if sub_value is not None:
                        sub_parts.append(f"{attr}:{sub_value}")
                sub_parts = sorted(sub_parts)
                query_parts.append("({})".format(" OR ".join(sub_parts)))
            else:
                value = _format_query_value(attr, value)
                if value is not None:
                    query_parts.append(f"{attr}:{value}")

        if raw:
            query_parts.append(raw)

        if area is not None:
            query_parts.append('footprint:"{}({})"'.format(area_relation, area))

        return " ".join(query_parts)

    def count(self, area=None, date=None, raw=None, area_relation="Intersects", **keywords):
        """Get the number of products matching a query.

        Accepted parameters are identical to :meth:`SentinelAPI.query()`.

        This is a significantly more efficient alternative to doing `len(api.query())`,
        which can take minutes to run for queries matching thousands of products.

        Returns
        -------
        int
            The number of products matching a query.
        """
        for kw in ["order_by", "limit", "offset"]:
            # Allow these function arguments to be included for compatibility with query(),
            # but ignore them.
            if kw in keywords:
                del keywords[kw]
        query = self.format_query(area, date, raw, area_relation, **keywords)
        _, total_count = self._load_query(query, limit=0)
        return total_count

    def _load_query(self, query, order_by=None, limit=None, offset=0):
        products, count = self._load_subquery(query, order_by, limit, offset)

        # repeat query until all results have been loaded
        max_offset = count
        if limit is not None:
            max_offset = min(count, offset + limit)
        if max_offset > offset + self.page_size:
            progress = self._tqdm(
                desc="Querying products",
                initial=self.page_size,
                total=max_offset - offset,
                unit=" products",
            )
            for new_offset in range(offset + self.page_size, max_offset, self.page_size):
                new_limit = limit
                if limit is not None:
                    new_limit = limit - new_offset + offset
                ret = self._load_subquery(query, order_by, new_limit, new_offset)[0]
                progress.update(len(ret))
                products += ret
            progress.close()

        return products, count

    def _load_subquery(self, query, order_by=None, limit=None, offset=0):
        # store last query (for testing)
        self._last_query = query
        self.logger.debug("Sub-query: offset=%s, limit=%s", offset, limit)

        # load query results
        url = self._format_url(order_by, limit, offset)
        # Unlike POST, DHuS only accepts latin1 charset in the GET params
        response = self.session.get(url, params={"q": query.encode("latin1")})
        _check_scihub_response(response, query_string=query)

        # store last status code (for testing)
        self._last_response = response

        # parse response content
        try:
            json_feed = response.json()["feed"]
            if "error" in json_feed:
                message = json_feed["error"]["message"]
                message = message.replace("org.apache.solr.search.SyntaxError: ", "")
                raise QuerySyntaxError(message, response)
            total_results = int(json_feed["opensearch:totalResults"])
        except (ValueError, KeyError):
            raise ServerError("API response not valid. JSON decoding failed.", response)

        products = json_feed.get("entry", [])
        # this verification is necessary because if the query returns only
        # one product, self.products will be a dict not a list
        if isinstance(products, dict):
            products = [products]

        return products, total_results

    def _format_url(self, order_by=None, limit=None, offset=0):
        if limit is None:
            limit = self.page_size
        limit = min(limit, self.page_size)
        url = "search?format=json&rows={}".format(limit)
        url += "&start={}".format(offset)
        if order_by:
            url += "&orderby={}".format(order_by)
        return urljoin(self.api_url, url)

    @staticmethod
    def to_geojson(products):
        """Return the products from a query response as a GeoJSON with the values in their
        appropriate Python types.
        """
        feature_list = []
        for i, (product_id, props) in enumerate(products.items()):
            props = props.copy()
            props["id"] = product_id
            poly = geomet.wkt.loads(props["footprint"])
            del props["footprint"]
            del props["gmlfootprint"]
            # Fix "'datetime' is not JSON serializable"
            for k, v in props.items():
                if isinstance(v, (date, datetime)):
                    props[k] = v.strftime("%Y-%m-%dT%H:%M:%S.%fZ")
            feature_list.append(geojson.Feature(geometry=poly, id=i, properties=props))
        return geojson.FeatureCollection(feature_list)

    @staticmethod
    def to_dataframe(products):
        """Return the products from a query response as a Pandas DataFrame
        with the values in their appropriate Python types.
        """
        try:
            import pandas as pd
        except ImportError:
            raise ImportError("to_dataframe requires the optional dependency Pandas.")

        return pd.DataFrame.from_dict(products, orient="index")

    @staticmethod
    def to_geodataframe(products):
        """Return the products from a query response as a GeoPandas GeoDataFrame
        with the values in their appropriate Python types.
        """
        try:
            import geopandas as gpd
            import shapely.wkt
        except ImportError:
            raise ImportError(
                "to_geodataframe requires the optional dependencies GeoPandas and Shapely."
            )

        crs = "EPSG:4326"  # WGS84
        if len(products) == 0:
            return gpd.GeoDataFrame(crs=crs, geometry=[])

        df = SentinelAPI.to_dataframe(products)
        geometry = [shapely.wkt.loads(fp) for fp in df["footprint"]]
        # remove useless columns
        df.drop(["footprint", "gmlfootprint"], axis=1, inplace=True)
        return gpd.GeoDataFrame(df, crs=crs, geometry=geometry)

    def get_product_odata(self, id, full=False):
        """Access OData API to get info about a product.

        Returns a dict containing the id, title, size, md5sum, date, footprint and download url
        of the product. The date field corresponds to the Start ContentDate value.

        If `full` is set to True, then the full, detailed metadata of the product is returned
        in addition to the above.

        Parameters
        ----------
        id : string
            The UUID of the product to query
        full : bool
            Whether to get the full metadata for the Product. False by default.

        Returns
        -------
        dict[str, Any]
            A dictionary with an item for each metadata attribute

        Notes
        -----
        For a full list of mappings between the OpenSearch (Solr) and OData attribute names
        see the following definition files:
        https://github.com/SentinelDataHub/DataHubSystem/blob/master/addon/sentinel-1/src/main/resources/META-INF/sentinel-1.owl
        https://github.com/SentinelDataHub/DataHubSystem/blob/master/addon/sentinel-2/src/main/resources/META-INF/sentinel-2.owl
        https://github.com/SentinelDataHub/DataHubSystem/blob/master/addon/sentinel-3/src/main/resources/META-INF/sentinel-3.owl
        """
        url = self._get_odata_url(id, "?$format=json")
        if full:
            url += "&$expand=Attributes"
        response = self.session.get(url)
        _check_scihub_response(response)
        values = _parse_odata_response(response.json()["d"])
        values["quicklook_url"] = self._get_odata_url(id, "/Products('Quicklook')/$value")
        return values

    def is_online(self, id):
        """Returns whether a product is online

        Parameters
        ----------
        id : string
            UUID of the product, e.g. 'a8dd0cfd-613e-45ce-868c-d79177b916ed'

        Returns
        -------
        bool
            True if online, False if in LTA

        See Also
        --------
        ``SentinelAPI.trigger_offline_retrieval()``
        """
        # Check https://scihub.copernicus.eu/userguide/ODataAPI#Products_entity for more information

        url = self._get_odata_url(id, "/Online/$value")
        r = self.session.get(url)
        _check_scihub_response(r)
        return r.json()

    def download(self, id, directory_path=".", checksum=True, **kwargs):
        """Download a product.

        Uses the filename on the server for the downloaded file, e.g.
        "S1A_EW_GRDH_1SDH_20141003T003840_20141003T003920_002658_002F54_4DD1.zip".

        Incomplete downloads are continued and complete files are skipped.

        Parameters
        ----------
        id : string
            UUID of the product, e.g. 'a8dd0cfd-613e-45ce-868c-d79177b916ed'
        directory_path : string, optional
            Where the file will be downloaded
        checksum : bool, optional
            If True, verify the downloaded file's integrity by checking its MD5 checksum.
            Throws InvalidChecksumError if the checksum does not match.
            Defaults to True.

        Returns
        -------
        product_info : dict
            Dictionary containing the product's info from get_product_info() as well as
            the path on disk.

        Raises
        ------
        InvalidChecksumError
            If the MD5 checksum does not match the checksum on the server.
        LTATriggered
            If the product has been archived and its retrieval was successfully triggered.
        LTAError
            If the product has been archived and its retrieval failed.

        .. versionchanged:: 1.0.0
           * Added ``**kwargs`` parameter to allow easier specialization of the :class:`SentinelAPI` class.
           * Now raises LTATriggered or LTAError if the product has been archived.
        """
        product_info = self.get_product_odata(id)
        filename = self._get_filename(product_info)
        path = Path(directory_path) / filename
        product_info["path"] = str(path)
        product_info["downloaded_bytes"] = 0

        self.logger.info("Downloading %s to %s", id, path)

        if path.exists():
            # We assume that the product has been downloaded and is complete
            return product_info

        # An incomplete download triggers the retrieval from the LTA if the product is not online
        is_online = not self.trigger_offline_retrieval(id)
        if not is_online:
            raise LTATriggered(id)

        self._download_outer(product_info, path, checksum)
        return product_info

    def _download_outer(self, product_info: Dict[str, Any], path: Path, verify_checksum: bool):
        # Use a temporary file for downloading
        temp_path = path.with_name(path.name + ".incomplete")
        skip_download = False
        if temp_path.exists():
            size = temp_path.stat().st_size
            if size > product_info["size"]:
                self.logger.warning(
                    "Existing incomplete file %s is larger than the expected final size"
                    " (%s vs %s bytes). Deleting it.",
                    str(temp_path),
                    size,
                    product_info["size"],
                )
                temp_path.unlink()
            elif size == product_info["size"]:
                if verify_checksum and not self._md5_compare(temp_path, product_info["md5"]):
                    # Log a warning since this should never happen
                    self.logger.warning(
                        "Existing incomplete file %s appears to be fully downloaded but "
                        "its checksum is incorrect. Deleting it.",
                        str(temp_path),
                    )
                    temp_path.unlink()
                else:
                    skip_download = True
            else:
                # continue downloading
                self.logger.info(
                    "Download will resume from existing incomplete file %s.", temp_path
                )
                pass
        if not skip_download:
            # Store the number of downloaded bytes for unit tests
            temp_path.parent.mkdir(parents=True, exist_ok=True)
            product_info["downloaded_bytes"] = self._download(
                product_info["url"], temp_path, product_info["size"]
            )
        # Check integrity with MD5 checksum
        if verify_checksum is True:
            if not self._md5_compare(temp_path, product_info["md5"]):
                temp_path.unlink()
                raise InvalidChecksumError("File corrupt: checksums do not match")
        # Download successful, rename the temporary file to its proper name
        shutil.move(temp_path, path)

    def _get_filename(self, product_info):
        if not product_info["Online"]:
            req = self.session.get(
                product_info["url"].replace("$value", "Attributes('Filename')/Value/$value")
            )
            _check_scihub_response(req, test_json=False)
            filename = req.text
            # This should cover all currently existing file types: .SAFE, .SEN3, .nc and .EOF
            filename = filename.replace(".SAFE", ".zip")
            filename = filename.replace(".SEN3", ".zip")
            return filename
        req = self.session.head(product_info["url"])
        _check_scihub_response(req, test_json=False)
        cd = req.headers.get("Content-Disposition")
        filename = cd.split("=", 1)[1].strip('"')
        return filename

    def trigger_offline_retrieval(self, uuid):
        """Triggers retrieval of an offline product.

        Trying to download an offline product triggers its retrieval from the long term archive.

        Parameters
        ----------
        uuid : string
            UUID of the product

        Returns
        -------
        bool
            True, if the product retrieval was successfully triggered.
            False, if the product is already online.

        Raises
        ------
        LTAError
            If the request was not accepted due to exceeded user quota or server overload.
        ServerError
            If an unexpected response was received from server.
        UnauthorizedError
            If the provided credentials were invalid.

        Notes
        -----
        https://scihub.copernicus.eu/userguide/LongTermArchive
        """
        # Request just a single byte to avoid accidental downloading of the whole product.
        # Requesting zero bytes results in NullPointerException in the server.
        r = self.session.get(self._get_download_url(uuid), headers={"Range": "bytes=0-1"})
        cause = r.headers.get("cause-message")
        # check https://scihub.copernicus.eu/userguide/LongTermArchive#HTTP_Status_codes
        if r.status_code in (200, 206):
            self.logger.debug("Product is online")
            return False
        elif r.status_code == 202:
            self.logger.debug("Accepted for retrieval")
            return True
        elif r.status_code == 403:
            msg = f"User quota exceeded: {cause}"
            self.logger.error(msg)
            raise LTAError(msg, r)
        elif r.status_code == 503:
            msg = f"Request not accepted: {cause}"
            self.logger.error(msg)
            raise LTAError(msg, r)
        elif r.status_code < 400:
            msg = f"Unexpected response {r.status_code}: {cause}"
            self.logger.error(msg)
            raise ServerError(msg, r)
        _check_scihub_response(r, test_json=False)

    def download_all(
        self,
        products,
        directory_path=".",
        max_attempts=10,
        checksum=True,
        n_concurrent_dl=2,
        lta_retry_delay=600,
        **kwargs
    ):
        """Download a list of products.

        Takes a list of product IDs as input. This means that the return value of query() can be
        passed directly to this method.

        File names on the server are used for the downloaded files, e.g.
        "S1A_EW_GRDH_1SDH_20141003T003840_20141003T003920_002658_002F54_4DD1.zip".

        In case of interruptions or other exceptions, downloading will restart from where it left
        off. Downloading is attempted at most max_attempts times to avoid getting stuck with
        unrecoverable errors.

        Parameters
        ----------
        products : list
            List of product IDs
        directory_path : string
            Directory where the downloaded files will be downloaded
        max_attempts : int, optional
            Number of allowed retries before giving up downloading a product. Defaults to 10.
        checksum : bool, optional
            If True, verify the downloaded files' integrity by checking its MD5 checksum.
            Throws InvalidChecksumError if the checksum does not match.
            Defaults to True.
        n_concurrent_dl : integer
            number of concurrent downloads
        lta_retry_delay : integer
            how long to wait between requests to the long term archive. Default is 600 seconds.
        **kwargs :
            additional parameters for the *download* method

        Raises
        ------
        Raises the most recent downloading exception if all downloads failed.

        Returns
        -------
        dict[string, dict]
            A dictionary containing the return value from download() for each successfully
            downloaded product.
        dict[string, dict]
            A dictionary containing the product information for products successfully
            triggered for retrieval from the long term archive but not downloaded.
        dict[string, dict]
            A dictionary containing the product information of products where either
            downloading or triggering failed. "exception" field with the exception info
            is included to the product info dict.


        .. versionchanged:: 0.15
           Added ``**kwargs`` parameter to allow easier specialization of the :class:`SentinelAPI` class.
        """

        product_ids = list(products)
        self.logger.info(
            "Will download %d products using %d workers", len(product_ids), n_concurrent_dl
        )

        product_infos = {}
        online_prods = {}
        offline_prods = {}
        for pid in self._tqdm(
            iterable=product_ids, desc="Fetching archival status", unit=" products"
        ):
            info = self.get_product_odata(pid)
            product_infos[pid] = info
            is_online = not self.trigger_offline_retrieval(pid)
            if is_online:
                online_prods[pid] = info
            else:
                offline_prods[pid] = info

        # Skip already downloaded files.
        # Although the download method also checks, we do not need to retrieve such
        # products from the LTA and use up our quota.
        downloaded_prods = {}
        for product_info in list(offline_prods.values()):
            filename = self._get_filename(product_info)
            path = Path(directory_path) / filename
            if path.exists():
                product_info["path"] = str(path)
                downloaded_prods[product_info["id"]] = product_info
                del offline_prods[product_info["id"]]
            else:
                self.logger.info("Product %s is in LTA.", product_info["id"])

        dl_tasks = {}
        retrieval_scheduled = {}

        # Two separate threadpools for downloading and triggering retrieval.
        # Otherwise triggering might take up all threads and nothing is downloaded.
        with concurrent.futures.ThreadPoolExecutor(max_workers=n_concurrent_dl) as dl_exec:
            # First all online products are downloaded. Subsequently, offline products that might
            # have become available in the meantime are requested.
            for product_info in itertools.chain(online_prods.values(), offline_prods.values()):
                dl_tasks[product_info["id"]] = dl_exec.submit(
                    self._download_online_retry,
                    product_info,
                    directory_path,
                    checksum,
                    max_attempts=max_attempts,
                    **kwargs
                )

            stop_event = threading.Event()
            trigger_thread = threading.Thread(
                target=self._trigger_offline_retrieval_until_stop,
                args=(offline_prods, stop_event, retrieval_scheduled, lta_retry_delay),
            )

            # launch in separate thread so that the as_completed loop is entered
            trigger_thread.start()

            for dl_task in concurrent.futures.as_completed(dl_tasks.values()):
                if not dl_task.cancelled() and not dl_task.exception():
                    product_info = dl_task.result()
                    if product_info is not None:
                        downloaded_prods[product_info["id"]] = product_info
                    # This else catches the first dl_task that did not complete because
                    # the product was not online and _download_online_retry returned None
                    else:
                        stop_event.set()
                        for task in dl_tasks.values():
                            task.cancel()

            # Wait for trigger_thread to finish. This could still place a product on the
            # retrieval_scheduled queue.
            trigger_thread.join()

        retrieval_scheduled = {
            pid: info
            for pid, info in retrieval_scheduled.items()
            if pid not in downloaded_prods.keys()
        }

        failed_prods = {}
        for pid, info in product_infos.items():
            if pid not in downloaded_prods and pid not in retrieval_scheduled:
                if dl_tasks[pid].cancelled():
                    info["exception"] = concurrent.futures.CancelledError()
                else:
                    info["exception"] = dl_tasks[pid].exception()
                failed_prods[pid] = info

        if len(failed_prods) == len(product_ids) and len(product_ids) > 0:
            exception = list(failed_prods.values())[0]["exception"]
            if exception is None:
                raise SentinelAPIError("Downloading all products failed for unknown reason")
            raise exception

        ResultTuple = namedtuple("ResultTuple", ["downloaded", "retrieval_triggered", "failed"])
        return ResultTuple(downloaded_prods, retrieval_scheduled, failed_prods)

    def _trigger_offline_retrieval_until_stop(
        self, product_infos, stop_event, retrieval_scheduled, retry_delay=600
    ):
        """Countinuously triggers retrieval of offline products

        This function is supposed to be called in a separate thread. By setting stop_event it can be stopped.

        Parameters
        ----------
        product_infos : dictionary
            Contains uuid of offline products as keys and their product information as values.
        stop_event: threading.Event
            If this event is set from another thread triggering from the LTA will stop
        retrieval_scheduled: dictionary
            Stores product information of triggered products. This can be accessed by other threads.
        retry_delay: integer
            After an unsuccessful triggering operation. Try again after this delay

        Notes
        -----
        https://scihub.copernicus.eu/userguide/LongTermArchive

        """

        for product_info in product_infos.values():
            while not stop_event.is_set():
                try:
                    triggered = self.trigger_offline_retrieval(product_info["id"])
                    if triggered:
                        self.logger.info("%s accepted for retrieval", product_info["id"])
                        retrieval_scheduled[product_info["id"]] = product_info
                        break
                except LTAError as e:
                    self.logger.info(
                        "Request for %s was not accepted: %s. Retrying in %d seconds",
                        product_info["id"],
                        e.msg,
                        retry_delay,
                    )
                    stop_event.wait(timeout=retry_delay)

    def _download_online_retry(
        self, product_info, directory_path=".", checksum=True, max_attempts=10, **kwargs
    ):
        """Thin wrapper around download with retrying and checking whether a product is online

        Parameters
        ----------

        product_info : dict
            Contains the product's info as returned by get_product_info()
        directory_path : string, optional
            Where the file will be downloaded
        checksum : bool, optional
            If True, verify the downloaded file's integrity by checking its MD5 checksum.
            Throws InvalidChecksumError if the checksum does not match.
            Defaults to True.
        max_attempts : int, optional
            Number of allowed retries before giving up downloading a product. Defaults to 10.

        Returns
        -------
        dict or None:
            Either dictionary containing the product's info or if the product is not online just None

        """
        if max_attempts <= 0:
            return

        id = product_info["id"]
        title = product_info["title"]
        is_online = not self.trigger_offline_retrieval(id)
        if not is_online:
            self.logger.info("%s is not online.", id)
            return

        self.logger.info("%s is online. Starting download", id)
        last_exception = None
        for cnt in range(max_attempts):
            try:
                return self.download(id, directory_path, checksum, **kwargs)
            except Exception as e:
                if isinstance(e, InvalidChecksumError):
                    self.logger.warning(
                        "Invalid checksum. The downloaded file for '%s' is corrupted.",
                        title,
                    )
                else:
                    self.logger.exception("There was an error downloading %s", title)
                self.logger.info("%d retries left", max_attempts - cnt - 1)
                last_exception = e
        self.logger.info("No retries left for %s. Terminating.", title)
        raise last_exception

    def download_all_quicklooks(self, products, directory_path=".", n_concurrent_dl=2):
        """Download quicklook for a list of products.

        Takes a dict of product IDs: product data as input. This means that the return value of
        query() can be passed directly to this method.

        File names on the server are used for the downloaded images, e.g.
        "S2A_MSIL1C_20200924T104031_N0209_R008_T35WMV_20200926T135405.jpeg".

        Parameters
        ----------
        products : dict
            Dict of product IDs, product data
        directory_path : string
            Directory where the downloaded images will be downloaded
        n_concurrent_dl : integer
            Number of concurrent downloads

        Returns
        -------
        dict[string, dict]
            A dictionary containing the return value from download_quicklook() for each
            successfully downloaded quicklook
        dict[string, dict]
            A dictionary containing the error of products where either
            quicklook was not available or it had an unexpected content type
        """

        self.logger.info("Will download %d quicklooks", len(products))

        downloaded_quicklooks = {}
        failed_quicklooks = {}

        with concurrent.futures.ThreadPoolExecutor(max_workers=n_concurrent_dl) as dl_exec:
            dl_tasks = {}
            for pid in products:
                future = dl_exec.submit(self.download_quicklook, pid, directory_path)
                dl_tasks[future] = pid

            completed_tasks = concurrent.futures.as_completed(dl_tasks)

            for future in completed_tasks:
                product_info = future.result()
                if product_info["error"] == "":
                    downloaded_quicklooks[dl_tasks[future]] = product_info
                else:
                    failed_quicklooks[dl_tasks[future]] = product_info["error"]

        ResultTuple = namedtuple("ResultTuple", ["downloaded", "failed"])
        return ResultTuple(downloaded_quicklooks, failed_quicklooks)

    def download_quicklook(self, id, directory_path="."):
        """Download a quicklook for a product.

        Uses the filename on the server for the downloaded image, e.g.
        "S1A_EW_GRDH_1SDH_20141003T003840_20141003T003920_002658_002F54_4DD1.jpeg".

        Complete images are skipped.

        Parameters
        ----------
        id : string
            UUID of the product, e.g. 'a8dd0cfd-613e-45ce-868c-d79177b916ed'
        directory_path : string, optional
            Where the image will be downloaded

        Returns
        -------
        quicklook_info : dict
            Dictionary containing the quicklooks's response headers as well as the path on disk.
        """
        product_info = self.get_product_odata(id)
        url = product_info["quicklook_url"]

        path = Path(directory_path) / "{}.jpeg".format(product_info["title"])
        product_info["path"] = str(path)
        product_info["downloaded_bytes"] = 0
        product_info["error"] = ""

        self.logger.info("Downloading quicklook %s to %s", product_info["title"], path)

        r = self.session.get(url)
        _check_scihub_response(r, test_json=False)

        product_info["quicklook_size"] = len(r.content)

        if path.exists():
            return product_info

        content_type = r.headers["content-type"]
        if content_type != "image/jpeg":
            product_info["error"] = "Quicklook is not jpeg but {}".format(content_type)

        if product_info["error"] == "":
            with open(path, "wb") as fp:
                fp.write(r.content)
                product_info["downloaded_bytes"] = len(r.content)

        return product_info

    @staticmethod
    def get_products_size(products):
        """Return the total file size in GB of all products in the OpenSearch response."""
        size_total = 0
        for title, props in products.items():
            size_product = props["size"]
            size_value = float(size_product.split(" ")[0])
            size_unit = str(size_product.split(" ")[1])
            if size_unit == "MB":
                size_value /= 1024.0
            if size_unit == "KB":
                size_value /= 1024.0 * 1024.0
            size_total += size_value
        return round(size_total, 2)

    @staticmethod
    def check_query_length(query):
        """Determine whether a query to the OpenSearch API is too long.

        The length of a query string is limited to approximately 3898 characters but
        any special characters (that is, not alphanumeric or -_.*) will take up more space.

        Parameters
        ----------
        query : str
            The query string

        Returns
        -------
        float
            Ratio of the query length to the maximum length
        """
        # The server uses the Java's URLEncoder implementation internally, which we are replicating here
        effective_length = len(quote_plus(query, safe="-_.*").replace("~", "%7E"))

        return effective_length / 3898

    def _query_names(self, names):
        """Find products by their names, e.g.
        S1A_EW_GRDH_1SDH_20141003T003840_20141003T003920_002658_002F54_4DD1.

        Note that duplicates exist on server, so multiple products can be returned for each name.

        Parameters
        ----------
        names : list[string]
            List of product names.

        Returns
        -------
        dict[string, dict[str, dict]]
            A dictionary mapping each name to a dictionary which contains the products with
            that name (with ID as the key).
        """
        products = {}
        for name in names:
            products.update(self.query(identifier=name))

        # Group the products
        output = OrderedDict((name, dict()) for name in names)
        for id, metadata in products.items():
            name = metadata["identifier"]
            output[name][id] = metadata

        return output

    def check_files(self, paths=None, ids=None, directory=None, delete=False):
        """Verify the integrity of product files on disk.

        Integrity is checked by comparing the size and checksum of the file with the respective
        values on the server.

        The input can be a list of products to check or a list of IDs and a directory.

        In cases where multiple products with different IDs exist on the server for given product
        name, the file is considered to be correct if any of them matches the file size and
        checksum. A warning is logged in such situations.

        The corrupt products' OData info is included in the return value to make it easier to
        re-download the products, if necessary.

        Parameters
        ----------
        paths : list[string]
            List of product file paths.
        ids : list[string]
            List of product IDs.
        directory : string
            Directory where the files are located, if checking based on product IDs.
        delete : bool
            Whether to delete corrupt products. Defaults to False.

        Returns
        -------
        dict[str, list[dict]]
            A dictionary listing the invalid or missing files. The dictionary maps the corrupt
            file paths to a list of OData dictionaries of matching products on the server (as
            returned by :meth:`SentinelAPI.get_product_odata()`).
        """
        if not ids and not paths:
            raise ValueError("Must provide either file paths or product IDs and a directory")
        if ids and not directory:
            raise ValueError("Directory value missing")
        if directory is not None:
            directory = Path(directory)
        paths = [Path(p) for p in paths] if paths else []
        ids = ids or []

        # Get product IDs corresponding to the files on disk
        names = []
        if paths:
            names = [p.stem for p in paths]
            result = self._query_names(names)
            for product_dicts in result.values():
                ids += list(product_dicts)
        names_from_paths = set(names)
        ids = set(ids)

        # Collect the OData information for each product
        # Product name -> list of matching odata dicts
        product_infos = defaultdict(list)
        for id in ids:
            odata = self.get_product_odata(id)
            name = odata["title"]
            product_infos[name].append(odata)

            # Collect
            if name not in names_from_paths:
                paths.append(directory / self._get_filename(odata))

        # Now go over the list of products and check them
        corrupt = {}
        for path in paths:
            name = path.stem

            if len(product_infos[name]) > 1:
                self.logger.warning("%s matches multiple products on server", path)

            if not path.exists():
                # We will consider missing files as corrupt also
                self.logger.info("%s does not exist on disk", path)
                corrupt[str(path)] = product_infos[name]
                continue

            is_fine = False
            for product_info in product_infos[name]:
                if path.stat().st_size == product_info["size"] and self._md5_compare(
                    path, product_info["md5"]
                ):
                    is_fine = True
                    break
            if not is_fine:
                self.logger.info("%s is corrupt", path)
                corrupt[str(path)] = product_infos[name]
                if delete:
                    path.unlink()

        return corrupt

    def _md5_compare(self, file_path, checksum, block_size=2 ** 13):
        """Compare a given MD5 checksum with one calculated from a file."""
        file_path = Path(file_path)
        file_size = file_path.stat().st_size
        with self._tqdm(
            desc="MD5 checksumming", total=file_size, unit="B", unit_scale=True
        ) as progress:
            md5 = hashlib.md5()
            with open(file_path, "rb") as f:
                while True:
                    block_data = f.read(block_size)
                    if not block_data:
                        break
                    md5.update(block_data)
                    progress.update(len(block_data))
            return md5.hexdigest().lower() == checksum.lower()

    def _download(self, url, path, file_size):
        headers = {}
        continuing = path.exists()
        if continuing:
            already_downloaded_bytes = path.stat().st_size
            headers = {"Range": "bytes={}-".format(already_downloaded_bytes)}
        else:
            already_downloaded_bytes = 0
        downloaded_bytes = 0
        with self.session.get(url, stream=True, headers=headers) as r, self._tqdm(
            desc="Downloading",
            total=file_size,
            unit="B",
            unit_scale=True,
            initial=already_downloaded_bytes,
        ) as progress:
            _check_scihub_response(r, test_json=False)
            chunk_size = 2 ** 20  # download in 1 MB chunks
            mode = "ab" if continuing else "wb"
            with open(path, mode) as f:
                for chunk in r.iter_content(chunk_size=chunk_size):
                    if chunk:  # filter out keep-alive new chunks
                        f.write(chunk)
                        progress.update(len(chunk))
                        downloaded_bytes += len(chunk)
            # Return the number of bytes downloaded
            return downloaded_bytes

    def _tqdm(self, **kwargs):
        """tqdm progressbar wrapper. May be overridden to customize progressbar behavior"""
        kwargs.update({"disable": not self.show_progressbars})
        return tqdm(**kwargs)

    def get_stream(self, id, **kwargs):
        """Exposes requests response ready to stream product to e.g. S3.

        Parameters
        ----------
        id : string
            UUID of the product, e.g. 'a8dd0cfd-613e-45ce-868c-d79177b916ed'
        **kwargs
            Any additional parameters for ``requests.get()``

        Raises
        ------
        LTATriggered
            If the product has been archived and its retrieval was successfully triggered.
        LTAError
            If the product has been archived and its retrieval failed.

        Returns
        -------
        requests.Response:
            Opened response object
        """
        is_online = not self.trigger_offline_retrieval(id)
        if not is_online:
            raise LTATriggered(id)
        r = self.session.get(self._get_download_url(id), stream=True, **kwargs)
        _check_scihub_response(r, test_json=False)
        return r

    def _get_odata_url(self, uuid, suffix=""):
        return self.api_url + f"odata/v1/Products('{uuid}')" + suffix

    def _get_download_url(self, uuid):
        return self._get_odata_url(uuid, "/$value")


def read_geojson(geojson_file):
    """Read a GeoJSON file into a GeoJSON object."""
    with open(geojson_file) as f:
        return geojson.load(f)


def geojson_to_wkt(geojson_obj, feature_number=0, decimals=4):
    """Convert a GeoJSON object to Well-Known Text. Intended for use with OpenSearch queries.

    In case of FeatureCollection, only one of the features is used (the first by default).
    3D points are converted to 2D.

    Parameters
    ----------
    geojson_obj : dict
        a GeoJSON object
    feature_number : int, optional
        Feature to extract polygon from (in case of MultiPolygon
        FeatureCollection), defaults to first feature
    decimals : int, optional
        Number of decimal figures after point to round coordinate to. Defaults to 4 (about 10
        meters).

    Returns
    -------
    str
        Well-Known Text string representation of the geometry
    """
    if "coordinates" in geojson_obj:
        geometry = geojson_obj
    elif "geometry" in geojson_obj:
        geometry = geojson_obj["geometry"]
    else:
        geometry = geojson_obj["features"][feature_number]["geometry"]

    def ensure_2d(geometry):
        if isinstance(geometry[0], (list, tuple)):
            return list(map(ensure_2d, geometry))
        else:
            return geometry[:2]

    def check_bounds(geometry):
        if isinstance(geometry[0], (list, tuple)):
            return list(map(check_bounds, geometry))
        else:
            if geometry[0] > 180 or geometry[0] < -180:
                raise ValueError("Longitude is out of bounds, check your JSON format or data")
            if geometry[1] > 90 or geometry[1] < -90:
                raise ValueError("Latitude is out of bounds, check your JSON format or data")

    # Discard z-coordinate, if it exists
    geometry["coordinates"] = ensure_2d(geometry["coordinates"])
    check_bounds(geometry["coordinates"])

    wkt = geomet.wkt.dumps(geometry, decimals=decimals)
    # Strip unnecessary spaces
    wkt = re.sub(r"(?<!\d) ", "", wkt)
    return wkt


def _format_query_value(attr, value):
    """Format the value of a Solr query parameter."""

    # Handle spaces by adding quotes around the string, if appropriate
    if isinstance(value, str):
        value = value.strip()
        if value == "":
            raise ValueError(f"Trying to filter '{attr}' with an empty string")
        # Handle strings surrounded by brackets specially to allow the user to make use of Solr syntax directly.
        # The string must not be quoted for that to work.
        if (
            not any(
                value.startswith(s[0]) and value.endswith(s[1])
                for s in ["[]", "{}", "//", "()", '""']
            )
            and "*" not in value
            and "?" not in value
        ):
            value = re.sub(r"\s", " ", value, re.M)
            value = f'"{value}"'

    # Handle date keywords
    # Keywords from https://github.com/SentinelDataHub/DataHubSystem/search?q=text/date+iso8601
    is_date_attr = attr.lower() in [
        "beginposition",
        "endposition",
        "date",
        "creationdate",
        "ingestiondate",
    ]
    if is_date_attr:
        # Automatically format date-type attributes
        if isinstance(value, str) and " TO " in value:
            # This is a string already formatted as a date interval,
            # e.g. '[NOW-1DAY TO NOW]'
            pass
        elif isinstance(value, (list, tuple)) and len(value) == 2:
            value = (format_query_date(value[0]), format_query_date(value[1]))
        else:
            raise ValueError(
                f"Date-type query parameter '{attr}' expects either a two-element tuple of "
                f"str or datetime objects or a '[<from> TO <to>]'-format string. Received {value}."
            )

    if isinstance(value, set):
        raise ValueError(f"Unexpected set-type value encountered with keyword '{attr}'")
    elif isinstance(value, (list, tuple)):
        # Handle value ranges
        if len(value) == 2:
            # Allow None or "*" to be used as an unlimited bound
            if any(x == "" for x in value):
                raise ValueError(f"Trying to filter '{attr}' with an empty string")
            value = ["*" if x in (None, "*") else f'"{x}"' for x in value]
            if value == ["*", "*"] or (is_date_attr and value == ["*", "NOW"]):
                # Drop this keyword if both sides are unbounded
                return
            value = "[{} TO {}]".format(*value)
        else:
            raise ValueError(
                f"Invalid number of elements in list. Expected 2, received {len(value)}"
            )
    return value


def format_query_date(in_date):
    r"""
    Format a date, datetime or a YYYYMMDD string input as YYYY-MM-DDThh:mm:ssZ
    or validate a date string as suitable for the full text search interface and return it.

    `None` will be converted to '\*', meaning an unlimited date bound in date ranges.

    Parameters
    ----------
    in_date : str or datetime or date or None
        Date to be formatted

    Returns
    -------
    str
        Formatted string

    Raises
    ------
    ValueError
        If the input date type is incorrect or passed date string is invalid
    """
    if in_date is None:
        return "*"
    if isinstance(in_date, (datetime, date)):
        return in_date.strftime("%Y-%m-%dT%H:%M:%SZ")
    elif not isinstance(in_date, str):
        raise ValueError("Expected a string or a datetime object. Received {}.".format(in_date))

    in_date = in_date.strip()
    if in_date == "*":
        # '*' can be used for one-sided range queries e.g. ingestiondate:[* TO NOW-1YEAR]
        return in_date

    # Reference: https://cwiki.apache.org/confluence/display/solr/Working+with+Dates

    # ISO-8601 date or NOW
    valid_date_pattern = r"^(?:\d{4}-\d\d-\d\dT\d\d:\d\d:\d\d(?:\.\d+)?Z|NOW)"
    # date arithmetic suffix is allowed
    units = r"(?:YEAR|MONTH|DAY|HOUR|MINUTE|SECOND)"
    valid_date_pattern += r"(?:[-+]\d+{}S?)*".format(units)
    # dates can be rounded to a unit of time
    # e.g. "NOW/DAY" for dates since 00:00 today
    valid_date_pattern += r"(?:/{}S?)*$".format(units)
    in_date = in_date.strip()
    if re.match(valid_date_pattern, in_date):
        return in_date

    try:
        return datetime.strptime(in_date, "%Y%m%d").strftime("%Y-%m-%dT%H:%M:%SZ")
    except ValueError:
        raise ValueError("Unsupported date value {}".format(in_date))


def _check_scihub_response(response, test_json=True, query_string=None):
    """Check that the response from server has status code 2xx and that the response is valid JSON."""
    # Prevent requests from needing to guess the encoding
    # SciHub appears to be using UTF-8 in all of their responses
    response.encoding = "utf-8"
    try:
        response.raise_for_status()
        if test_json:
            response.json()
    except (requests.HTTPError, ValueError):
        msg = None
        try:
            msg = response.json()["error"]["message"]["value"]
        except Exception:
            try:
                msg = response.headers["cause-message"]
            except Exception:
                if not response.text.lstrip().startswith("{"):
                    try:
                        h = html2text.HTML2Text()
                        h.ignore_images = True
                        h.ignore_anchors = True
                        msg = h.handle(response.text).strip()
                    except Exception:
                        pass

        if msg is None:
            raise ServerError("Invalid API response", response)
        elif response.status_code == 401:
            msg = "Invalid user name or password"
            if "apihub.copernicus.eu/apihub" in response.url:
                msg += (
                    ". Note that account creation and password changes may take up to a week "
                    "to propagate to the 'https://apihub.copernicus.eu/apihub/' API URL you are using. "
                    "Consider switching to 'https://scihub.copernicus.eu/dhus/' instead in the mean time."
                )
            raise UnauthorizedError(msg, response)
        elif "Request Entity Too Large" in msg or "Request-URI Too Long" in msg:
            msg = "Server was unable to process the query due to its excessive length"
            if query_string is not None:
                length = SentinelAPI.check_query_length(query_string)
                msg += (
                    " (approximately {:.3}x times the maximum allowed). "
                    "Consider using SentinelAPI.check_query_length() for "
                    "client-side validation of the query string length.".format(length)
                )
            raise QueryLengthError(msg, response)
        elif "Invalid key" in msg:
            msg = msg.split(" : ", 1)[-1]
            raise InvalidKeyError(msg, response)
        elif 500 <= response.status_code < 600 or msg:
            # 5xx: Server Error
            raise ServerError(msg, response)
        else:
            raise SentinelAPIError(msg, response)


def _format_order_by(order_by):
    if not order_by or not order_by.strip():
        return None
    output = []
    for part in order_by.split(","):
        part = part.strip()
        dir = " asc"
        if part[0] == "+":
            part = part[1:]
        elif part[0] == "-":
            dir = " desc"
            part = part[1:]
        if not part or not part.isalnum():
            raise ValueError("Invalid order by value ({})".format(order_by))
        output.append(part + dir)
    return ",".join(output)


def _parse_gml_footprint(geometry_str):
    # workaround for https://github.com/sentinelsat/sentinelsat/issues/286
    if geometry_str is None:  # pragma: no cover
        return None
    geometry_xml = ET.fromstring(geometry_str)
    poly_coords_str = (
        geometry_xml.find("{http://www.opengis.net/gml}outerBoundaryIs")
        .find("{http://www.opengis.net/gml}LinearRing")
        .findtext("{http://www.opengis.net/gml}coordinates")
    )
    poly_coords = (coord.split(",")[::-1] for coord in poly_coords_str.split(" "))
    coord_string = ",".join(" ".join(coord) for coord in poly_coords)
    return "POLYGON(({}))".format(coord_string)


def _parse_iso_date(content):
    if "." in content:
        return datetime.strptime(content, "%Y-%m-%dT%H:%M:%S.%fZ")
    else:
        return datetime.strptime(content, "%Y-%m-%dT%H:%M:%SZ")


def _parse_odata_timestamp(in_date):
    """Convert the timestamp received from OData JSON API to a datetime object."""
    timestamp = int(in_date.replace("/Date(", "").replace(")/", ""))
    seconds = timestamp // 1000
    ms = timestamp % 1000
    return datetime.utcfromtimestamp(seconds) + timedelta(milliseconds=ms)


def _parse_opensearch_response(products):
    """Convert a query response to a dictionary.

    The resulting dictionary structure is {<product id>: {<property>: <value>}}.
    The property values are converted to their respective Python types unless `parse_values`
    is set to `False`.
    """

    converters = {"date": _parse_iso_date, "int": int, "long": int, "float": float, "double": float}

    # Keep the string type by default
    def default_converter(x):
        return x

    output = OrderedDict()
    for prod in products:
        product_dict = {}
        prod_id = prod["id"]
        output[prod_id] = product_dict
        for key in prod:
            if key == "id":
                continue
            if isinstance(prod[key], str):
                product_dict[key] = prod[key]
            else:
                properties = prod[key]
                if isinstance(properties, dict):
                    properties = [properties]
                if key == "link":
                    for p in properties:
                        name = "link"
                        if "rel" in p:
                            name = "link_" + p["rel"]
                        product_dict[name] = p["href"]
                else:
                    f = converters.get(key, default_converter)
                    for p in properties:
                        try:
                            product_dict[p["name"]] = f(p["content"])
                        except KeyError:
                            # Sentinel-3 has one element 'arr'
                            # which violates the name:content convention
                            product_dict[p["name"]] = f(p["str"])
    return output


def _parse_odata_response(product):
    output = {
        "id": product["Id"],
        "title": product["Name"],
        "size": int(product["ContentLength"]),
        product["Checksum"]["Algorithm"].lower(): product["Checksum"]["Value"],
        "date": _parse_odata_timestamp(product["ContentDate"]["Start"]),
        "footprint": _parse_gml_footprint(product["ContentGeometry"]),
        "url": product["__metadata"]["media_src"],
        "Online": product.get("Online", True),
        "Creation Date": _parse_odata_timestamp(product["CreationDate"]),
        "Ingestion Date": _parse_odata_timestamp(product["IngestionDate"]),
    }
    # Parse the extended metadata, if provided
    converters = [int, float, _parse_iso_date]
    for attr in product["Attributes"].get("results", []):
        value = attr["Value"]
        for f in converters:
            try:
                value = f(attr["Value"])
                break
            except ValueError:
                pass
        output[attr["Name"]] = value
    return output


def placename_to_wkt(place_name):
    """Geocodes the place name to rectangular bounding extents using Nominatim API and
       returns the corresponding 'ENVELOPE' form Well-Known-Text.

    Parameters
    ----------
    place_name : str
        the query to geocode

    Raises
    ------
    ValueError
        If no matches to the place name were found.

    Returns
    -------
    wkt_envelope : str
        Bounding box of the location as an 'ENVELOPE(minX, maxX, maxY, minY)' WKT string.
    info : Dict[str, any]
        Matched location's metadata returned by Nominatim.
    """
    rqst = requests.get(
        "https://nominatim.openstreetmap.org/search",
        params={"q": place_name, "format": "geojson"},
        headers={"User-Agent": "sentinelsat/" + sentinelsat_version},
    )
    rqst.raise_for_status()
    features = rqst.json()["features"]
    if len(features) == 0:
        raise ValueError('Unable to find a matching location for "{}"'.format(place_name))
    # Get the First result's bounding box and description.
    feature = features[0]
    minX, minY, maxX, maxY = feature["bbox"]
    # ENVELOPE is a non-standard WKT format supported by Solr
    # https://lucene.apache.org/solr/guide/6_6/spatial-search.html#SpatialSearch-BBoxField
    wkt_envelope = "ENVELOPE({}, {}, {}, {})".format(minX, maxX, maxY, minY)
    info = feature["properties"]
    info["bbox"] = feature["bbox"]
    return wkt_envelope, info


def is_wkt(possible_wkt):
    pattern = r"^((MULTI)?(POINT|LINESTRING|POLYGON)|GEOMETRYCOLLECTION|ENVELOPE)\s*\(.+\)$"
    return re.match(pattern, possible_wkt.strip().upper()) is not None



import re
from datetime import datetime
from functools import reduce
from os import environ
from os.path import join, isfile, dirname, abspath, exists

import pytest
import yaml
from pytest_socket import disable_socket
from vcr import VCR

from sentinelsat import SentinelAPI, geojson_to_wkt, read_geojson, SentinelProductsAPI
from sentinelsat.sentinel import _parse_odata_response
from .custom_serializer import BinaryContentSerializer

TESTS_DIR = dirname(abspath(__file__))
FIXTURES_DIR = join(TESTS_DIR, "fixtures")
CASSETTE_DIR = join(FIXTURES_DIR, "vcr_cassettes")


def pytest_runtest_setup(item):
    markers = {mark.name for mark in item.iter_markers()}
    if "pandas" in markers:
        pytest.importorskip("pandas")
    if "geopandas" in markers:
        pytest.importorskip("geopandas")

    if not markers.intersection({"scihub", "fast", "mock_api"}):
        pytest.fail("The test is missing a 'scihub', 'fast' or 'mock_api' marker")


def scrub_request(request):
    for header in ("Authorization", "Set-Cookie", "Cookie"):
        if header in request.headers:
            del request.headers[header]
    return request


def scrub_response(response):
    ignore = {
        x.lower()
        for x in [
            "Authorization",
            "Set-Cookie",
            "Cookie",
            "Date",
            "Expires",
            "Transfer-Encoding",
            "last-modified",
        ]
    }
    for header in list(response["headers"]):
        if (
            header.lower() in ignore
            or header.lower().startswith("access-control")
            or header.lower().startswith("x-")
        ):
            del response["headers"][header]
    return response


def scrub_string(string, replacement=b""):
    """Scrub a string from a VCR response body string"""

    def before_record_response(response):
        len_before = len(response["body"]["string"])
        response["body"]["string"] = re.sub(string, replacement, response["body"]["string"])
        len_diff = len(response["body"]["string"]) - len_before
        if "content-length" in response["headers"]:
            response["headers"]["content-length"] = [
                str(int(response["headers"]["content-length"][0]) + len_diff)
            ]
        return response

    return before_record_response


def chain(*funcs):
    def chained_call(arg):
        return reduce(lambda x, f: f(x), funcs, arg)

    return chained_call


# Configure pytest-vcr
@pytest.fixture(scope="module")
def vcr(vcr):
    def range_header_matcher(r1, r2):
        return r1.headers.get("Range", "") == r2.headers.get("Range", "")

    vcr.cassette_library_dir = CASSETTE_DIR
    vcr.path_transformer = VCR.ensure_suffix(".yaml")
    vcr.filter_headers = ["Set-Cookie"]
    vcr.before_record_request = scrub_request
    vcr.before_record_response = chain(
        scrub_response,
        scrub_string(rb"Request done in \S+ seconds.", b"Request done in ... seconds."),
        scrub_string(rb'"updated":"[^"]+"', b'"updated":"..."'),
        scrub_string(rb'totalResults":"\d{4,}"', b'totalResults":"10000"'),
        scrub_string(rb"of \d{4,} total results", b"of 10000 total results"),
        scrub_string(rb"&start=\d{4,}&rows=0", b"&start=10000&rows=0"),
    )
    vcr.decode_compressed_response = True
    vcr.register_serializer("custom", BinaryContentSerializer(CASSETTE_DIR))
    vcr.serializer = "custom"
    vcr.register_matcher("range_header", range_header_matcher)
    vcr.match_on = ["method", "range_header", "host", "port", "path", "query", "body"]
    return vcr


@pytest.fixture(scope="session")
def credentials(request):
    # local tests require environment variables `DHUS_USER` and `DHUS_PASSWORD`
    # for Travis CI they are set as encrypted environment variables and stored
    record_mode = request.config.getoption("--vcr-record")
    disable_vcr = request.config.getoption("--disable-vcr")
    if record_mode in ["none", None] and not disable_vcr:
        # Using VCR.py cassettes for pre-recorded query playback
        # Any network traffic will raise an exception
        disable_socket()
    elif "DHUS_USER" not in environ or "DHUS_PASSWORD" not in environ:
        raise ValueError(
            "Credentials must be set when --vcr-record is not none or --disable-vcr is used. "
            "Please set DHUS_USER and DHUS_PASSWORD environment variables."
        )

    return [environ.get("DHUS_USER"), environ.get("DHUS_PASSWORD")]


@pytest.fixture(scope="session")
def api_kwargs(credentials):
    user, password = credentials
    return dict(user=user, password=password, api_url="https://apihub.copernicus.eu/apihub/")


@pytest.fixture
def api(api_kwargs):
    return SentinelAPI(**api_kwargs)


@pytest.fixture
def products_api(api_kwargs):
    return SentinelProductsAPI(**api_kwargs)


@pytest.fixture(scope="session")
def fixture_path():
    return lambda filename: join(FIXTURES_DIR, filename)


@pytest.fixture(scope="session")
def read_fixture_file(fixture_path):
    def read_func(filename, mode="r"):
        with open(fixture_path(filename), mode) as f:
            return f.read()

    return read_func


@pytest.fixture(scope="session")
def read_yaml(fixture_path, read_fixture_file):
    def read_or_store(filename, result):
        path = fixture_path(filename)
        if not exists(path):
            # Store the expected result for future if the fixture file is missing
            with open(path, "w") as f:
                yaml.safe_dump(result, f)
        return yaml.safe_load(read_fixture_file(filename))

    return read_or_store


@pytest.fixture(scope="session")
def geojson_path():
    path = join(FIXTURES_DIR, "map.geojson")
    assert isfile(path)
    return path


@pytest.fixture(scope="session")
def geojson_string():
    string = """{
  "type": "FeatureCollection",
  "features": [
    {
      "type": "Feature",
      "properties": {},
      "geometry": {
        "type": "Polygon",
        "coordinates": [
          [
            [
              -66.26953125,
              -8.05922962720018
            ],
            [
              -66.26953125,
              0.7031073524364909
            ],
            [
              -57.30468749999999,
              0.7031073524364909
            ],
            [
              -57.30468749999999,
              -8.05922962720018
            ],
            [
              -66.26953125,
              -8.05922962720018
            ]
          ]
        ]
      }
    }
  ]
}"""
    return string


@pytest.fixture(scope="session")
def wkt_string():
    string = (
        "POLYGON((-78.046875 46.377254205100286,-75.76171874999999 43.32517767999295,-71.279296875 "
        "46.55886030311717,-78.046875 46.377254205100286))"
    )
    return string


@pytest.fixture(scope="session")
def test_wkt(geojson_path):
    return geojson_to_wkt(read_geojson(geojson_path))


@pytest.fixture(scope="module")
def products(api_kwargs, vcr, test_wkt):
    """A fixture for tests that need some non-specific set of products as input."""
    with vcr.use_cassette("products_fixture", decode_compressed_response=False):
        api = SentinelAPI(**api_kwargs)
        products = api.query(test_wkt, ("20151219", "20151228"))
    assert len(products) > 20
    return products


@pytest.fixture(scope="module")
def raw_products(api_kwargs, vcr, test_wkt):
    """A fixture for tests that need some non-specific set of products in the form of a raw response as input."""
    with vcr.use_cassette("products_fixture", decode_compressed_response=False):
        api = SentinelAPI(**api_kwargs)
        raw_products = api._load_query(api.format_query(test_wkt, ("20151219", "20151228")))[0]
    return raw_products


def _get_smallest(api_kwargs, cassette, online, n=3):
    api = SentinelAPI(**api_kwargs)
    filter = " and ".join(
        [
            "Online eq {}".format("true" if online else "false"),
            # limit time range for tests stability and much faster OData querying
            "CreationDate lt datetime'2016-01-01T00:00:00.000'",
            # needed for SentinelProductsAPI tests, some S5 apparently do not include manifest.safe
            "not startswith(Name, 'S3')",
            # don't want zero-length products for testing, plus their handling has server-side bugs
            "ContentLength ne 0",
        ]
    )
    url = "{}odata/v1/Products?$format=json&$top={}&$filter={}".format(
        api_kwargs["api_url"], n + 3, filter
    )
    if online:
        # archived products do not have the correct ContentLength value,
        # so use an arbitrary subset of the data and skip orderby to speed up the query
        url += "&$orderby=ContentLength"
    with cassette:
        r = api.session.get(url)
    odata = [_parse_odata_response(x) for x in r.json()["d"]["results"]]
    # Drop products that appear to be broken
    blacklist = ["S2A_MSIL2A_20190528T113321_N0212_R080_T29UPB_20190528T142130"]
    odata = [x for x in odata if x["title"] not in blacklist]
    odata = odata[:n]
    assert len(odata) == n
    return odata


def _get_quicklook(api_kwargs, cassette):
    api = SentinelAPI(**api_kwargs)
    ids = [
        "6b126ea4-fe27-440c-9a5c-686f386b7291",
        "1a9401bc-6986-4707-b38d-f6c29ca58c00",
        "54e6c4ad-6f4e-4fbf-b163-1719f60bfaeb",
    ]
    with cassette:
        odata = [api.get_product_odata(x) for x in ids]
    return odata


@pytest.fixture(scope="module")
def smallest_online_products(api_kwargs, vcr):
    return _get_smallest(api_kwargs, vcr.use_cassette("smallest_online_products"), online=True)


@pytest.fixture(scope="module")
def smallest_archived_products(api_kwargs, vcr):
    n = 3
    api = SentinelAPI(**api_kwargs)
    # Find some small and old products expecting them to be archived due to age.
    # Can't use the OData API for this as we do for the online products
    # because the ContentLength value there is not match the true product size.
    odatas = []
    with vcr.use_cassette("smallest_archived_products"):
        products = api.query(date=(None, "20170101"), size="/.+KB/", limit=10)
        for uuid in products:
            odata = api.get_product_odata(uuid)
            if not odata["Online"]:
                odatas.append(odata)
                if len(odatas) == n:
                    break
        assert len(odatas) == n
    return odatas


@pytest.fixture(scope="module")
def quicklook_products(api_kwargs, vcr):
    return _get_quicklook(api_kwargs, vcr.use_cassette("quicklook_products"))


@pytest.fixture(scope="session")
def small_query():
    return dict(
        area="POLYGON((0 0,1 1,0 1,0 0))", date=(datetime(2015, 1, 1), datetime(2015, 1, 2))
    )


@pytest.fixture(scope="session")
def large_query():
    return dict(
        area="POLYGON((0 0,0 10,10 10,10 0,0 0))",
        date=(datetime(2015, 12, 1), datetime(2015, 12, 31)),
    )



class SentinelAPIError(Exception):
    """Invalid response from DataHub. Base class for more specific exceptions.

    Attributes
    ----------
    msg: str
        The error message.
    response: requests.Response
        The response from the server as a `requests.Response` object.
    """

    def __init__(self, msg="", response=None):
        self.msg = msg
        self.response = response

    def __str__(self):
        if self.response is None:
            return self.msg
        if self.response.reason:
            reason = " " + self.response.reason
        else:
            reason = ""
        return "HTTP status {}{}: {}".format(
            self.response.status_code,
            reason,
            ("\n" if "\n" in self.msg else "") + self.msg,
        )


class LTAError(SentinelAPIError):
    """Error raised when retrieving a product from the Long Term Archive"""

    pass


class LTATriggered(SentinelAPIError):
    """Download failed due to product being in the Long Term Archive and retrieval from LTA was triggered"""

    def __init__(self, uuid):
        self.uuid = uuid
        super().__init__(
            f"Product {uuid} is not online. Triggered retrieval from the Long Term Archive."
        )


class ServerError(SentinelAPIError):
    """Error raised when the server responded in an unexpected manner, typically due to undergoing maintenance"""

    pass


class UnauthorizedError(SentinelAPIError):
    """Error raised when attempting to retrieve a product with incorrect credentials"""

    def __str__(self):
        return self.msg


class QuerySyntaxError(SentinelAPIError, SyntaxError):
    """Error raised when the query string could not be parsed on the server side"""

    def __init__(self, msg, response):
        SentinelAPIError.__init__(self, msg, response)
        SyntaxError.__init__(self, msg)

    def __str__(self):
        return self.msg


class QueryLengthError(SentinelAPIError):
    """Error raised when the query string length was excessively long"""

    def __str__(self):
        return self.msg


class InvalidKeyError(SentinelAPIError, KeyError):
    """Error raised when product with given key was not found on the server"""

    def __init__(self, msg, response):
        SentinelAPIError.__init__(self, msg, response)
        KeyError.__init__(self, msg)

    def __str__(self):
        return self.msg


class InvalidChecksumError(Exception):
    """MD5 checksum of a local file does not match the one from the server."""

    pass



import json
import logging
import math
import os
import re

import click
import geojson as gj
import requests.utils

from sentinelsat import __version__ as sentinelsat_version

from sentinelsat.sentinel import (
    geojson_to_wkt,
    read_geojson,
    placename_to_wkt,
    is_wkt,
)

from sentinelsat.products import SentinelProductsAPI, make_path_filter


json_parse_exception = json.decoder.JSONDecodeError

logger = logging.getLogger("sentinelsat")


def _set_logger_handler(level="INFO"):
    logger.setLevel(level)
    h = logging.StreamHandler()
    h.setLevel(level)
    h.setFormatter(logging.Formatter("%(message)s"))
    logger.addHandler(h)


def validate_query_param(ctx, param, kwargs):
    for kwarg in kwargs:
        if not re.match(r"\w+=.+", kwarg):
            raise click.BadParameter("must have the format 'keyword=value'")
    return kwargs


@click.command(context_settings=dict(help_option_names=["-h", "--help"]))
@click.option(
    "--user",
    "-u",
    envvar="DHUS_USER",
    default=None,
    help="Username (or environment variable DHUS_USER is set)",
)
@click.option(
    "--password",
    "-p",
    envvar="DHUS_PASSWORD",
    default=None,
    help="Password (or environment variable DHUS_PASSWORD is set)",
)
@click.option(
    "--url",
    default="https://apihub.copernicus.eu/apihub/",
    envvar="DHUS_URL",
    help="""Define API URL. Default URL is
        'https://apihub.copernicus.eu/apihub/' (or environment variable DHUS_URL).
        """,
)
@click.option(
    "--start",
    "-s",
    default=None,
    help="Start date of the query in the format YYYYMMDD or an expression like NOW-1DAY.",
)
@click.option(
    "--end",
    "-e",
    default=None,
    help="End date of the query.",
)
@click.option(
    "--geometry",
    "-g",
    type=str,
    help="Search area geometry as GeoJSON file, a GeoJSON string, or a WKT string. ",
)
@click.option(
    "--uuid",
    multiple=True,
    help="Select a specific product UUID. Can be set more than once.",
)
@click.option(
    "--name",
    multiple=True,
    help="Select specific product(s) by filename. Supports wildcards. Can be set more than once.",
)
@click.option(
    "--sentinel",
    type=click.Choice(["1", "2", "3", "5"]),
    help="Limit search to a Sentinel satellite (constellation)",
)
@click.option(
    "--instrument",
    type=click.Choice(["MSI", "SAR-C SAR", "SLSTR", "OLCI", "SRAL"]),
    help="Limit search to a specific instrument on a Sentinel satellite.",
)
@click.option(
    "--producttype", type=str, default=None, help="Limit search to a Sentinel product type."
)
@click.option(
    "-c",
    "--cloud",
    type=int,
    help="Maximum cloud cover in percent. (requires --sentinel to be 2 or 3)",
)
@click.option(
    "-o",
    "--order-by",
    help="Comma-separated list of keywords to order the result by. "
    "Prefix keywords with '-' for descending order.",
)
@click.option(
    "-l", "--limit", type=int, help="Maximum number of results to return. Defaults to no limit."
)
@click.option("--download", "-d", is_flag=True, help="Download all results of the query.")
@click.option(
    "--quicklook",
    is_flag=True,
    help="""Download quicklook of a product.""",
)
@click.option(
    "--path",
    type=click.Path(exists=True),
    default=".",
    help="Set the path where the files will be saved.",
)
@click.option(
    "--query",
    "-q",
    multiple=True,
    callback=validate_query_param,
    help="""Extra search keywords you want to use in the query. 
        Example: '-q producttype=GRD -q polarisationmode=HH'.
        Repeated keywords are interpreted as an "or" expression.
        """,
)
@click.option(
    "--location",
    type=str,
    help="Return only products overlapping with the bounding box of given location, "
    "e.g. 'Berlin', 'Germany' or '52.393974, 13.066955'.",
)
@click.option(
    "--footprints",
    type=click.File(mode="w", encoding="utf8", lazy=True),
    help="""Create a GeoJSON file at the provided path with footprints
    and metadata of the returned products. Set to '-' for stdout.
    """,
)
@click.option(
    "--timeout",
    type=float,
    default=60,
    help="How long to wait for a DataHub response (in seconds, default 60 sec).",
)
@click.option(
    "--debug",
    is_flag=True,
    help="Print debug log messages.",
)
@click.option(
    "--include-pattern",
    default=None,
    help="""Glob pattern to filter files (within each product) to be downloaded.
    """,
)
@click.option(
    "--exclude-pattern",
    default=None,
    help="""Glob pattern to filter files (within each product) to be excluded
    from the downloaded.
    """,
)
@click.option(
    "--gnss",
    is_flag=True,
    help="""Use the "https://scihub.copernicus.eu/gnss" end-point
    for orbit data query and download.
    """,
)
@click.option("--info", is_flag=True, is_eager=True, help="Displays the DHuS version used")
@click.version_option(version=sentinelsat_version, prog_name="sentinelsat")
def cli(
    user,
    password,
    geometry,
    start,
    end,
    uuid,
    name,
    download,
    quicklook,
    sentinel,
    producttype,
    instrument,
    cloud,
    footprints,
    path,
    query,
    url,
    order_by,
    location,
    limit,
    timeout,
    debug,
    include_pattern,
    exclude_pattern,
    gnss,
    info,
):
    """Search for Sentinel products and, optionally, download all the results
    and/or create a geojson file with the search result footprints.
    Beyond your Copernicus Open Access Hub user and password, you must pass a geojson file
    containing the geometry of the area you want to search for or the UUIDs of the products. If you
    don't specify the start and end dates, it will search in the last 24 hours.
    """

    if footprints and footprints.name == "-":
        _set_logger_handler("WARNING")
    elif debug:
        _set_logger_handler("DEBUG")
    else:
        _set_logger_handler("INFO")

    if info:
        api = SentinelProductsAPI(None, None, url, timeout=timeout)
        ctx = click.get_current_context()
        click.echo("DHuS version: " + api.dhus_version)
        ctx.exit()

    if include_pattern is not None and exclude_pattern is not None:
        raise click.UsageError(
            "--include-pattern and --exclude-pattern cannot be specified together."
        )
    elif include_pattern is not None:
        nodefilter = make_path_filter(include_pattern)
    elif exclude_pattern is not None:
        nodefilter = make_path_filter(exclude_pattern, exclude=True)
    else:
        nodefilter = None

    if gnss:
        url = "https://scihub.copernicus.eu/gnss/"
        user = "gnssguest"
        password = "gnssguest"

    if user is None or password is None:
        try:
            user, password = requests.utils.get_netrc_auth(url)
        except TypeError:
            pass

    if user is None or password is None:
        raise click.UsageError(
            "Missing --user and --password. Please see docs "
            "for environment variables and .netrc support."
        )

    api = SentinelProductsAPI(user, password, url, timeout=timeout)

    search_kwargs = {}
    if sentinel:
        search_kwargs["platformname"] = "Sentinel-" + sentinel

    if instrument:
        search_kwargs["instrumentshortname"] = instrument

    if producttype:
        search_kwargs["producttype"] = producttype

    if cloud:
        if sentinel not in ["2", "3"]:
            logger.error("Cloud cover is only supported for Sentinel 2 and 3.")
            exit(1)
        search_kwargs["cloudcoverpercentage"] = (0, cloud)

    if len(name) > 0:
        search_kwargs["identifier"] = set(name)

    if len(uuid) > 0:
        search_kwargs["uuid"] = set(uuid)

    if len(query) > 0:
        for kwarg in query:
            key, value = kwarg.split("=", 1)
            if key in search_kwargs:
                if isinstance(search_kwargs[key], set):
                    search_kwargs[key].add(value)
                else:
                    search_kwargs[key] = {search_kwargs[key], value}
            else:
                search_kwargs[key] = value

    if location is not None:
        wkt, info = placename_to_wkt(location)
        minX, minY, maxX, maxY = info["bbox"]
        r = 6371  # average radius, km
        extent_east = r * math.radians(maxX - minX) * math.cos(math.radians((minY + maxY) / 2))
        extent_north = r * math.radians(maxY - minY)
        logger.info(
            "Querying location: '%s' with %.1f x %.1f km, %f, %f to %f, %f bounding box",
            info["display_name"],
            extent_north,
            extent_east,
            minY,
            minX,
            maxY,
            maxX,
        )
        search_kwargs["area"] = wkt

    if geometry is not None:
        # check if the value is an existing path
        if os.path.exists(geometry):
            search_kwargs["area"] = geojson_to_wkt(read_geojson(geometry))
        # check if the value is a GeoJSON
        else:
            if geometry.startswith("{"):
                try:
                    geometry = json.loads(geometry)
                    search_kwargs["area"] = geojson_to_wkt(geometry)
                except json_parse_exception:
                    raise click.UsageError(
                        "geometry string starts with '{' but is not a valid GeoJSON."
                    )
            # check if the value is a WKT
            elif is_wkt(geometry):
                search_kwargs["area"] = geometry
            else:
                raise click.UsageError(
                    "The geometry input is neither a GeoJSON file with a valid path, "
                    "a GeoJSON String nor a WKT string."
                )

    products = api.query(date=(start, end), order_by=order_by, limit=limit, **search_kwargs)

    if footprints is not None:
        footprints_geojson = api.to_geojson(products)
        gj.dump(footprints_geojson, footprints)
        footprints.close()

    if quicklook:
        downloaded_quicklooks, failed_quicklooks = api.download_all_quicklooks(products, path)
        if failed_quicklooks:
            api.logger.warning(
                "Some quicklooks failed: %s out of %s", len(failed_quicklooks), len(products)
            )

    if download is True:
        product_infos, triggered, failed_downloads = api.download_all(
            products, path, nodefilter=nodefilter
        )
        if len(failed_downloads) > 0:
            with open(os.path.join(path, "corrupt_scenes.txt"), "w") as outfile:
                for failed_id in failed_downloads:
                    outfile.write("{} : {}\n".format(failed_id, products[failed_id]["title"]))
    else:
        for product_id, props in products.items():
            logger.info("Product %s - %s", product_id, props["summary"])
        logger.info("---")
        logger.info(
            "%s scenes found with a total size of %.2f GB",
            len(products),
            api.get_products_size(products),
        )



"""
Product downloading related tests.

There are two minor issues to keep in mind when recording unit tests VCRs.

1. Between calls a formerly offline product can become available, if the previous call triggered its LTA retrieval.
2. dhus and apihub have different md5 hashes for products with the same UUID.

"""
import os
import shutil

import py.path
import pytest
import requests_mock
from flaky import flaky

from sentinelsat import SentinelAPI, make_path_filter
from sentinelsat.exceptions import InvalidChecksumError, InvalidKeyError, LTAError, ServerError


@pytest.mark.fast
@pytest.mark.parametrize(
    "api_url, dhus_url",
    [
        ("https://apihub.copernicus.eu/apihub/", "https://scihub.copernicus.eu/dhus/"),
        ("https://colhub.met.no/", "https://colhub.met.no/"),
        ("https://finhub.nsdc.fmi.fi/", "https://finhub.nsdc.fmi.fi/"),
    ],
)
def test_api2dhus_url(api_url, dhus_url):
    api = SentinelAPI("mock_user", "mock_password")
    assert api._api2dhus_url(api_url) == dhus_url


@pytest.mark.mock_api
@pytest.mark.parametrize(
    "dhus_url, version",
    [
        # version numbers retrieved on January 31st, 2020
        ("https://scihub.copernicus.eu/dhus", "2.4.1"),
        ("https://colhub.met.no", "0.13.4-22"),
        ("https://finhub.nsdc.fmi.fi", "0.13.4-21-1"),
    ],
)
def test_dhus_version(dhus_url, version):
    api = SentinelAPI("mock_user", "mock_password", api_url=dhus_url)
    request_url = dhus_url + "/api/stub/version"
    with requests_mock.mock() as rqst:
        rqst.get(request_url, json={"value": version})
        assert api.dhus_version == version


@pytest.mark.mock_api
@pytest.mark.parametrize(
    "http_status_code, expected_result",
    [
        # Note: the HTTP status codes have slightly more specific meanings in the LTA API.
        # OK - already online
        (200, False),
        (206, False),
        # Accepted for retrieval - the product offline product will be retrieved from the LTA.
        (202, True),
    ],
)
def test_trigger_lta_success(http_status_code, expected_result):
    api = SentinelAPI("mock_user", "mock_password")
    uuid = "8df46c9e-a20c-43db-a19a-4240c2ed3b8b"

    with requests_mock.mock() as rqst:
        rqst.get(api._get_download_url(uuid), status_code=http_status_code)
        assert api.trigger_offline_retrieval(uuid) is expected_result


@pytest.mark.mock_api
@pytest.mark.parametrize(
    "http_status_code, exception",
    [
        # Note: the HTTP status codes have slightly more specific meanings in the LTA API.
        # Forbidden - user has exceeded their offline product retrieval quota.
        (403, LTAError),
        # Service Unavailable - request refused since the service is busy handling other requests.
        (503, LTAError),
        # Internal Server Error - attempted to download a sub-element of an offline product.
        (500, ServerError),
        (555, ServerError),
        (333, ServerError),
    ],
)
def test_trigger_lta_failed(http_status_code, exception):
    api = SentinelAPI("mock_user", "mock_password")
    uuid = "8df46c9e-a20c-43db-a19a-4240c2ed3b8b"

    with requests_mock.mock() as rqst:
        rqst.get(api._get_download_url(uuid), status_code=http_status_code)
        with pytest.raises(exception):
            api.trigger_offline_retrieval(uuid)


@pytest.mark.vcr
@pytest.mark.scihub
def test_download(api, tmpdir, smallest_online_products):
    product = smallest_online_products[0]
    uuid = product["id"]
    title = product["title"]
    expected_path = tmpdir.join(title + ".zip")
    tempfile_path = tmpdir.join(title + ".zip.incomplete")

    # Download normally
    product_info = api.download(uuid, str(tmpdir), checksum=True)
    assert expected_path.samefile(product_info["path"])
    assert not tempfile_path.check(exists=1)
    assert product_info["title"] == title
    assert product_info["size"] == expected_path.size()
    assert product_info["downloaded_bytes"] == expected_path.size()

    hash = expected_path.computehash("md5")
    modification_time = expected_path.mtime()
    expected_product_info = product_info

    # File exists, expect nothing to happen
    product_info = api.download(uuid, str(tmpdir))
    assert not tempfile_path.check(exists=1)
    assert expected_path.mtime() == modification_time
    expected_product_info["downloaded_bytes"] = 0
    assert product_info == expected_product_info

    # Create invalid but full-sized tempfile, expect re-download
    expected_path.move(tempfile_path)
    with tempfile_path.open("wb") as f:
        f.seek(expected_product_info["size"] - 1)
        f.write(b"\0")
    assert tempfile_path.computehash("md5") != hash
    product_info = api.download(uuid, str(tmpdir))
    assert expected_path.check(exists=1, file=1)
    assert expected_path.computehash("md5") == hash
    expected_product_info["downloaded_bytes"] = expected_product_info["size"]
    assert product_info == expected_product_info

    # Create invalid tempfile, without checksum check
    # Expect continued download and no exception
    dummy_content = b"aaaaaaaaaaaaaaaaaaaaaaaaa"
    with tempfile_path.open("wb") as f:
        f.write(dummy_content)
    expected_path.remove()
    product_info = api.download(uuid, str(tmpdir), checksum=False)
    assert not tempfile_path.check(exists=1)
    assert expected_path.check(exists=1, file=1)
    assert expected_path.computehash("md5") != hash
    expected_product_info["downloaded_bytes"] = expected_product_info["size"] - len(dummy_content)
    assert product_info == expected_product_info

    # Create invalid tempfile, with checksum check
    # Expect continued download and exception raised
    dummy_content = b"aaaaaaaaaaaaaaaaaaaaaaaaa"
    with tempfile_path.open("wb") as f:
        f.write(dummy_content)
    expected_path.remove()
    with pytest.raises(InvalidChecksumError):
        api.download(uuid, str(tmpdir), checksum=True)
    assert not tempfile_path.check(exists=1)
    assert not expected_path.check(exists=1, file=1)

    tmpdir.remove()


@pytest.mark.vcr(allow_playback_repeats=True)
@pytest.mark.scihub
def test_download_all(api, tmpdir, smallest_online_products):
    ids = [product["id"] for product in smallest_online_products]

    # Download normally
    product_infos, triggered, failed_downloads = api.download_all(
        ids, str(tmpdir), max_attempts=1, n_concurrent_dl=1
    )
    assert len(failed_downloads) == 0
    assert len(triggered) == 0
    assert len(product_infos) == len(ids)
    for product_id, product_info in product_infos.items():
        pypath = py.path.local(product_info["path"])
        assert pypath.check(exists=1, file=1)
        assert pypath.purebasename in product_info["title"]
        assert pypath.size() == product_info["size"]


@pytest.mark.vcr(allow_playback_repeats=True)
@pytest.mark.scihub
def test_download_all_one_fail(api, tmpdir, smallest_online_products):
    ids = [product["id"] for product in smallest_online_products]

    # Force one download to fail
    id = ids[0]
    with requests_mock.mock(real_http=True) as rqst:
        url = "https://apihub.copernicus.eu/apihub/odata/v1/Products('%s')?$format=json" % id
        json = api.session.get(url).json()
        json["d"]["Checksum"]["Value"] = "00000000000000000000000000000000"
        rqst.get(url, json=json)
        product_infos, triggered, failed_downloads = api.download_all(
            ids, str(tmpdir), max_attempts=1, n_concurrent_dl=1, checksum=True
        )
        exceptions = {k: v["exception"] for k, v in failed_downloads.items()}
        for e in exceptions.values():
            if not isinstance(e, InvalidChecksumError):
                raise e from None
        assert sorted(failed_downloads) == [ids[0]], exceptions
        assert type(list(exceptions.values())[0]) == InvalidChecksumError, exceptions
        assert triggered == {}
        assert sorted(list(product_infos) + list(failed_downloads)) == sorted(ids)
        assert id in failed_downloads

    tmpdir.remove()


# VCR.py can't handle multi-threading correctly
# https://github.com/kevin1024/vcrpy/issues/212
@flaky(max_runs=3, min_passes=2)
@pytest.mark.vcr(allow_playback_repeats=True)
@pytest.mark.scihub
def test_download_all_lta(api, tmpdir, smallest_online_products, smallest_archived_products):
    archived_ids = [x["id"] for x in smallest_archived_products]
    online_ids = [x["id"] for x in smallest_online_products]
    ids = archived_ids[:1] + online_ids[:2]
    product_infos, triggered, failed_downloads = api.download_all(
        ids, str(tmpdir), max_attempts=1, n_concurrent_dl=1
    )
    exceptions = {k: v["exception"] for k, v in failed_downloads.items()}
    assert len(failed_downloads) == 0, exceptions
    assert len(triggered) == 1
    assert len(product_infos) == len(ids) - len(failed_downloads) - len(triggered)
    assert all(x["Online"] is False for x in triggered.values())

    # test downloaded products
    for product_id, product_info in product_infos.items():
        pypath = py.path.local(product_info["path"])
        assert pypath.check(exists=1, file=1)
        assert pypath.purebasename in product_info["title"]
        assert pypath.size() == product_info["size"]

    tmpdir.remove()


@pytest.mark.vcr
@pytest.mark.scihub
def test_download_invalid_id(api):
    uuid = "1f62a176-c980-41dc-xxxx-c735d660c910"
    with pytest.raises(InvalidKeyError) as excinfo:
        api.download(uuid)
    assert "Invalid key" in str(excinfo.value)


@pytest.mark.vcr
@pytest.mark.scihub
def test_download_quicklook(api, tmpdir, quicklook_products):
    uuid = quicklook_products[0]["id"]
    filename = quicklook_products[0]["title"]
    expected_path = tmpdir.join(filename + ".jpeg")
    # import pdb; pdb.set_trace()
    # Download normally
    quicklook_info = api.download_quicklook(uuid, str(tmpdir))
    assert expected_path.samefile(quicklook_info["path"])
    assert quicklook_info["title"] == filename
    assert quicklook_info["quicklook_size"] == expected_path.size()
    assert quicklook_info["downloaded_bytes"] == expected_path.size()

    modification_time = expected_path.mtime()
    expected_quicklook_info = quicklook_info

    # File exists, expect nothing to happen
    quicklook_info = api.download_quicklook(uuid, str(tmpdir))
    assert expected_path.mtime() == modification_time
    expected_quicklook_info["downloaded_bytes"] = 0
    assert quicklook_info["quicklook_size"] == expected_path.size()
    assert quicklook_info == expected_quicklook_info

    tmpdir.remove()


@pytest.mark.vcr(allow_playback_repeats=True)
@pytest.mark.scihub
def test_download_all_quicklooks(api, tmpdir, quicklook_products):
    ids = [product["id"] for product in quicklook_products]

    # Download normally
    downloaded_quicklooks, failed_quicklooks = api.download_all_quicklooks(
        ids, str(tmpdir), n_concurrent_dl=1
    )
    assert len(failed_quicklooks) == 0
    assert len(downloaded_quicklooks) == len(ids)
    for product_id, product_info in downloaded_quicklooks.items():
        pypath = py.path.local(product_info["path"])
        assert pypath.check(exists=1, file=1)
        assert pypath.purebasename in product_info["title"]
        assert pypath.size() == product_info["quicklook_size"]

    tmpdir.remove()


@pytest.mark.vcr(allow_playback_repeats=True)
@pytest.mark.scihub
def test_download_all_quicklooks_one_fail(api, tmpdir, quicklook_products):
    ids = [product["id"] for product in quicklook_products]

    # Force one download to fail
    id = ids[0]
    with requests_mock.mock(real_http=True) as rqst:
        url = f"https://apihub.copernicus.eu/apihub/odata/v1/Products('{id}')/Products('Quicklook')/$value"
        headers = api.session.get(url).headers
        headers["content-type"] = "image/xxxx"
        rqst.get(url, headers=headers)
        downloaded_quicklooks, failed_quicklooks = api.download_all_quicklooks(
            ids, str(tmpdir), n_concurrent_dl=1
        )
        assert len(failed_quicklooks) == 1
        assert len(downloaded_quicklooks) + len(failed_quicklooks) == len(ids)
        assert id in failed_quicklooks

    tmpdir.remove()


@pytest.mark.vcr
@pytest.mark.scihub
def test_download_quicklook_invalid_id(api):
    uuid = "1f62a176-c980-41dc-xxxx-c735d660c910"
    with pytest.raises(InvalidKeyError) as excinfo:
        api.download_quicklook(uuid)
    assert "Invalid key" in str(excinfo.value)


@pytest.mark.vcr
@pytest.mark.scihub
def test_get_stream(api, tmpdir, smallest_online_products):
    product_info = smallest_online_products[0]
    uuid = product_info["id"]
    filename = product_info["title"]
    expected_path = tmpdir.join(filename + ".zip")

    response = api.get_stream(uuid)
    assert product_info["title"] == filename
    with open(expected_path, "wb") as f:
        shutil.copyfileobj(response.raw, f)

    assert product_info["size"] == expected_path.size()
    assert api._md5_compare(expected_path, product_info["md5"])

    tmpdir.remove()


@pytest.mark.vcr
@pytest.mark.scihub
def test_download_product_nodes(products_api, tmpdir, smallest_online_products):
    uuid = smallest_online_products[0]["id"]
    product_dir = smallest_online_products[0]["title"] + ".SAFE"
    expected_path = tmpdir.join(product_dir)

    nodefilter = make_path_filter("*preview/*.kml")
    product_info = products_api.download(uuid, str(tmpdir), checksum=True, nodefilter=nodefilter)

    assert os.path.normpath(product_info["node_path"]) == product_dir
    assert expected_path.exists()

    assert set(product_info["nodes"]) == {"./manifest.safe", "./preview/map-overlay.kml"}

    assert tmpdir.join(product_dir, "manifest.safe").check()
    assert tmpdir.join(product_dir, "preview", "map-overlay.kml").check()

    assert not tmpdir.join(product_dir, "manifest.safe" + ".incomplete").check()
    assert not tmpdir.join(product_dir, "preview", "map-overlay.kml" + ".incomplete").check()

    tmpdir.remove()



import hashlib

import pytest
import requests
import requests_mock

from sentinelsat import SentinelAPI
from sentinelsat.exceptions import SentinelAPIError, QuerySyntaxError, InvalidKeyError
from sentinelsat.sentinel import _parse_opensearch_response


@pytest.mark.fast
def test_checksumming_progressbars(capsys, fixture_path):
    api = SentinelAPI("mock_user", "mock_password")
    testfile_md5 = hashlib.md5()
    true_path = fixture_path("expected_search_footprints_s1.geojson")
    with open(true_path, "rb") as testfile:
        testfile_md5.update(testfile.read())
        real_md5 = testfile_md5.hexdigest()

    assert api._md5_compare(true_path, real_md5) is True
    out, err = capsys.readouterr()
    assert "checksumming" in err
    api = SentinelAPI("mock_user", "mock_password", show_progressbars=False)
    assert api._md5_compare(fixture_path("map.geojson"), real_md5) is False
    out, err = capsys.readouterr()
    assert out == ""
    assert "checksumming" not in err


@pytest.mark.vcr
@pytest.mark.scihub
def test_unicode_support(api):
    # DHuS only accepts latin1 charset in the GET params
    with pytest.raises(UnicodeEncodeError):
        api.count(raw="٩(●̮̮̃•̃)۶:")

    # check that the allowed non-ASCII chars are at least understood correctly by DHuS
    test_str = "õäöü\xff("
    with pytest.raises(QuerySyntaxError) as excinfo:
        api.count(raw=test_str)
    assert test_str in str(excinfo.value)

    with pytest.raises(InvalidKeyError) as excinfo:
        api.get_product_odata(test_str)
    assert test_str in excinfo.value.response.json()["error"]["message"]["value"]
    assert test_str in str(excinfo.value)


@pytest.mark.mock_api
def test_scihub_unresponsive(small_query):
    timeout_connect = 6
    timeout_read = 6.6
    timeout = (timeout_connect, timeout_read)

    api = SentinelAPI("mock_user", "mock_password", timeout=timeout)

    with requests_mock.mock() as rqst:
        rqst.request(requests_mock.ANY, requests_mock.ANY, exc=requests.exceptions.ConnectTimeout)
        with pytest.raises(requests.exceptions.ConnectTimeout):
            api.query(**small_query)

        with pytest.raises(requests.exceptions.ConnectTimeout):
            api.get_product_odata("8df46c9e-a20c-43db-a19a-4240c2ed3b8b")

        with pytest.raises(requests.exceptions.ConnectTimeout):
            api.download("8df46c9e-a20c-43db-a19a-4240c2ed3b8b")

        with pytest.raises(requests.exceptions.ConnectTimeout):
            api.download_all(["8df46c9e-a20c-43db-a19a-4240c2ed3b8b"])

    with requests_mock.mock() as rqst:
        rqst.request(requests_mock.ANY, requests_mock.ANY, exc=requests.exceptions.ReadTimeout)
        with pytest.raises(requests.exceptions.ReadTimeout):
            api.query(**small_query)

        with pytest.raises(requests.exceptions.ReadTimeout):
            api.get_product_odata("8df46c9e-a20c-43db-a19a-4240c2ed3b8b")

        with pytest.raises(requests.exceptions.ReadTimeout):
            api.download("8df46c9e-a20c-43db-a19a-4240c2ed3b8b")

        with pytest.raises(requests.exceptions.ReadTimeout):
            api.download_all(["8df46c9e-a20c-43db-a19a-4240c2ed3b8b"])


@pytest.mark.mock_api
def test_get_products_invalid_json(test_wkt):
    api = SentinelAPI("mock_user", "mock_password")
    with requests_mock.mock() as rqst:
        rqst.get(
            "https://apihub.copernicus.eu/apihub/search",
            text="{Invalid JSON response",
            status_code=200,
        )
        with pytest.raises(SentinelAPIError) as excinfo:
            api.query(area=test_wkt, date=("20151219", "20151228"), platformname="Sentinel-2")
        assert excinfo.value.msg == "Invalid API response"


@pytest.mark.scihub
def test_get_products_size(api, vcr, products):
    assert SentinelAPI.get_products_size(products) == 75.4

    # load a new very small query
    with vcr.use_cassette("test_get_products_size"):
        products = api.query(
            raw="S1A_WV_OCN__2SSH_20150603T092625_20150603T093332_006207_008194_521E"
        )
    assert len(products) > 0
    # Rounded to zero
    assert SentinelAPI.get_products_size(products) == 0


@pytest.mark.scihub
def test_response_to_dict(raw_products):
    dictionary = _parse_opensearch_response(raw_products)
    # check the type
    assert isinstance(dictionary, dict)
    # check if dictionary has id key
    assert "bd1204f7-71ba-4b67-a5f4-df16fbb10138" in dictionary
    props = dictionary["bd1204f7-71ba-4b67-a5f4-df16fbb10138"]
    expected_title = "S2A_MSIL1C_20151223T142942_N0201_R053_T20MNC_20151223T143132"
    assert props["title"] == expected_title


@pytest.mark.vcr
@pytest.mark.scihub
def test_check_existing(api, tmpdir, smallest_online_products, smallest_archived_products):
    ids = [product["id"] for product in smallest_online_products]
    names = [product["title"] for product in smallest_online_products]
    paths = [tmpdir.join(fn + ".zip") for fn in names]
    path_strings = list(map(str, paths))

    # Init files used for testing
    api.download(ids[0], str(tmpdir))
    # File #1: complete and correct
    assert paths[0].check(exists=1, file=1)
    # File #2: complete but incorrect
    with paths[1].open("wb") as f:
        size = 130102
        f.seek(size - 1)
        f.write(b"\0")
    # File #3: incomplete
    dummy_content = b"aaaaaaaaaaaaaaaaaaaaaaaaa"
    with paths[2].open("wb") as f:
        f.write(dummy_content)
    assert paths[2].check(exists=1, file=1)

    # Test
    expected = {str(paths[1]), str(paths[2])}

    def check_result(result, expected_existing):
        assert set(result) == expected
        assert result[str(paths[1])][0]["id"] == ids[1]
        assert result[str(paths[2])][0]["id"] == ids[2]
        assert [p.check(exists=1, file=1) for p in paths] == expected_existing

    result = api.check_files(ids=ids, directory=str(tmpdir))
    check_result(result, [True, True, True])

    result = api.check_files(paths=path_strings)
    check_result(result, [True, True, True])

    result = api.check_files(paths=path_strings, delete=True)
    check_result(result, [True, False, False])

    missing_file = str(tmpdir.join(smallest_archived_products[0]["title"] + ".zip"))
    result = api.check_files(paths=[missing_file])
    assert set(result) == {missing_file}
    assert result[missing_file][0]["id"] == smallest_archived_products[0]["id"]

    with pytest.raises(ValueError):
        api.check_files(ids=ids)

    with pytest.raises(ValueError):
        api.check_files()

    tmpdir.remove()



__version__ = "1.0.1"

# Import for backwards-compatibility
from . import sentinel
from .exceptions import (
    SentinelAPIError,
    LTAError,
    LTATriggered,
    ServerError,
    InvalidKeyError,
    QueryLengthError,
    QuerySyntaxError,
    UnauthorizedError,
    InvalidChecksumError,
)
from .sentinel import (
    SentinelAPI,
    format_query_date,
    geojson_to_wkt,
    read_geojson,
    placename_to_wkt,
)

from .products import (
    SentinelProductsAPI,
    make_path_filter,
    make_size_filter,
    all_nodes_filter,
)



"""
Tests for functionality related to the OpenSearch API of SciHub (https://apihub.copernicus.eu/apihub/search?...)
"""
from contextlib import contextmanager
from datetime import datetime, date, timedelta

import pytest
import requests_mock
from requests.exceptions import (
    StreamConsumedError,
    ContentDecodingError,
    InvalidProxyURL,
    InvalidHeader,
    TooManyRedirects,
    ReadTimeout,
    SSLError,
)

from sentinelsat import format_query_date, SentinelAPI
from sentinelsat.exceptions import QuerySyntaxError, QueryLengthError, UnauthorizedError
from sentinelsat.sentinel import _format_order_by


@pytest.mark.vcr
@pytest.mark.scihub
def test_format_date(api):
    assert format_query_date(datetime(2015, 1, 1)) == "2015-01-01T00:00:00Z"
    assert format_query_date(date(2015, 1, 1)) == "2015-01-01T00:00:00Z"
    assert format_query_date("2015-01-01T00:00:00Z") == "2015-01-01T00:00:00Z"
    assert format_query_date("20150101") == "2015-01-01T00:00:00Z"
    assert format_query_date(" NOW ") == "NOW"
    assert format_query_date(None) == "*"

    for date_str in (
        "NOW",
        "NOW-1DAY",
        "NOW-1DAYS",
        "NOW-500DAY",
        "NOW-500DAYS",
        "NOW-2MONTH",
        "NOW-2MONTHS",
        "NOW-20MINUTE",
        "NOW-20MINUTES",
        "NOW+10HOUR",
        "2015-01-01T00:00:00Z+1DAY",
        "NOW+3MONTHS-7DAYS/DAYS",
        "*",
    ):
        assert format_query_date(date_str) == date_str
        api.query(raw="ingestiondate:[{} TO *]".format(date_str), limit=0)

    for date_str in (
        "NOW - 1HOUR",
        "NOW -   1HOURS",
        "NOW-1 HOURS",
        "NOW-1",
        "NOW-",
        "**",
        "+",
        "-",
    ):
        with pytest.raises(ValueError):
            format_query_date(date_str)
        with pytest.raises(QuerySyntaxError):
            api.query(raw="ingestiondate:[{} TO *]".format(date_str), limit=0)


@pytest.mark.vcr
@pytest.mark.scihub
def test_SentinelAPI_connection(api, small_query):
    api.query(**small_query)
    assert api._last_query == (
        'beginPosition:["2015-01-01T00:00:00Z" TO "2015-01-02T00:00:00Z"] '
        'footprint:"Intersects(POLYGON((0 0,1 1,0 1,0 0)))"'
    )
    assert api._last_response.status_code == 200


@pytest.mark.vcr
@pytest.mark.scihub
def test_SentinelAPI_wrong_credentials(small_query):
    api = SentinelAPI("wrong_user", "wrong_password")

    @contextmanager
    def assert_exception():
        with pytest.raises(UnauthorizedError) as excinfo:
            yield
        assert excinfo.value.response.status_code == 401
        assert str(excinfo.value).startswith("Invalid user name or password")

    with assert_exception():
        api.query(**small_query)
    with assert_exception():
        api.get_product_odata("8df46c9e-a20c-43db-a19a-4240c2ed3b8b")
    with assert_exception():
        api.download("8df46c9e-a20c-43db-a19a-4240c2ed3b8b")
    with assert_exception():
        api.download_all(["8df46c9e-a20c-43db-a19a-4240c2ed3b8b"])


@pytest.mark.vcr
@pytest.mark.scihub
@pytest.mark.parametrize(
    "exception",
    [
        StreamConsumedError,
        ContentDecodingError,
        InvalidProxyURL,
        InvalidHeader,
        TooManyRedirects,
        ReadTimeout,
        SSLError,
    ],
)
def test_requests_error(exception, api, small_query):
    """non-HTTP errors originating from requests should be raised directly without translating them"""
    with requests_mock.mock() as rqst:
        rqst.register_uri(requests_mock.ANY, requests_mock.ANY, exc=exception)
        with pytest.raises(exception):
            api.query(**small_query)

        with pytest.raises(exception):
            api.get_product_odata("8df46c9e-a20c-43db-a19a-4240c2ed3b8b")

        with pytest.raises(exception):
            api.download("8df46c9e-a20c-43db-a19a-4240c2ed3b8b")

        with pytest.raises(exception):
            api.download_all(["8df46c9e-a20c-43db-a19a-4240c2ed3b8b"])


@pytest.mark.fast
def test_api_query_format():
    wkt = "POLYGON((0 0,1 1,0 1,0 0))"

    now = datetime.now()
    last_24h = format_query_date(now - timedelta(hours=24))
    query = SentinelAPI.format_query(wkt, (last_24h, now))
    assert (
        query
        == 'beginPosition:["{}" TO "{}"] '.format(last_24h, format_query_date(now))
        + 'footprint:"Intersects(POLYGON((0 0,1 1,0 1,0 0)))"'
    )

    query = SentinelAPI.format_query(wkt, date=(last_24h, "NOW"), producttype="SLC", raw="IW")
    assert (
        query
        == 'beginPosition:["{}" TO "NOW"] '.format(format_query_date(last_24h))
        + 'producttype:"SLC" IW footprint:"Intersects(POLYGON((0 0,1 1,0 1,0 0)))"'
    )

    query = SentinelAPI.format_query(wkt, producttype="SLC", raw="IW")
    assert query == 'producttype:"SLC" IW footprint:"Intersects(POLYGON((0 0,1 1,0 1,0 0)))"'

    query = SentinelAPI.format_query(area=None, date=None)
    assert query == ""

    query = SentinelAPI.format_query()
    assert query == ""

    query = SentinelAPI.format_query(raw="test")
    assert query == "test"


@pytest.mark.fast
def test_api_query_format_with_duplicates():
    with pytest.raises(ValueError) as excinfo:
        SentinelAPI.format_query(date=("NOW-1DAY", "NOW"), beginPosition=("NOW-3DAY", "NOW"))
    assert "duplicate" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        SentinelAPI.format_query(
            ingestiondate=("NOW-1DAY", "NOW"), ingestionDate=("NOW-3DAY", "NOW")
        )
    assert "duplicate" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        SentinelAPI.format_query(area="POINT(0, 0)", footprint="POINT(0, 0)")
    assert "duplicate" in str(excinfo.value)


@pytest.mark.fast
def test_api_query_format_ranges():
    query = SentinelAPI.format_query(cloudcoverpercentage=(0, 30))
    assert query == 'cloudcoverpercentage:["0" TO "30"]'

    query = SentinelAPI.format_query(cloudcoverpercentage=[0, 30])
    assert query == 'cloudcoverpercentage:["0" TO "30"]'

    query = SentinelAPI.format_query(cloudcoverpercentage=[None, 30])
    assert query == 'cloudcoverpercentage:[* TO "30"]'

    query = SentinelAPI.format_query(orbitnumber=(16302, None))
    assert query == 'orbitnumber:["16302" TO *]'

    query = SentinelAPI.format_query(orbitnumber=(16302, "*"))
    assert query == 'orbitnumber:["16302" TO *]'

    for value in [(None, None), ("*", None), (None, "*"), ("*", "*")]:
        query = SentinelAPI.format_query(orbitnumber=value)
        assert query == ""

    with pytest.raises(ValueError):
        SentinelAPI.format_query(cloudcoverpercentage=[])

    with pytest.raises(ValueError):
        SentinelAPI.format_query(cloudcoverpercentage=[0])

    with pytest.raises(ValueError):
        SentinelAPI.format_query(cloudcoverpercentage=[0, 1, 2])

    with pytest.raises(ValueError):
        SentinelAPI.format_query(cloudcoverpercentage=[0, ""])


@pytest.mark.fast
def test_api_query_format_sets():
    query = SentinelAPI.format_query(orbitnumber={16301, 16302, 16303})
    assert query == "(orbitnumber:16301 OR orbitnumber:16302 OR orbitnumber:16303)"

    query = SentinelAPI.format_query(orbitnumber=set())
    assert query == ""

    query = SentinelAPI.format_query(ingestiondate={(date(2017, 1, 1), "20170203")})
    assert query == '(ingestiondate:["2017-01-01T00:00:00Z" TO "2017-02-03T00:00:00Z"])'

    with pytest.raises(ValueError):
        SentinelAPI.format_query(ingestiondate={})
    with pytest.raises(ValueError):
        SentinelAPI.format_query(ingestiondate={""})
    with pytest.raises(ValueError):
        SentinelAPI.format_query(ingestiondate={date(2017, 1, 1), "20170203"})


@pytest.mark.fast
def test_api_query_format_dates():
    query = SentinelAPI.format_query(ingestiondate=("NOW-1DAY", "NOW"))
    assert query == 'ingestiondate:["NOW-1DAY" TO "NOW"]'

    query = SentinelAPI.format_query(ingestiondate=(date(2017, 1, 1), "20170203"))
    assert query == 'ingestiondate:["2017-01-01T00:00:00Z" TO "2017-02-03T00:00:00Z"]'

    query = SentinelAPI.format_query(ingestiondate="[NOW-1DAY TO NOW]")
    assert query == "ingestiondate:[NOW-1DAY TO NOW]"

    query = SentinelAPI.format_query(ingestiondate=[None, "NOW"])
    assert query == 'ingestiondate:[* TO "NOW"]'

    for value in [(None, None), ("*", None), (None, "*"), ("*", "*")]:
        query = SentinelAPI.format_query(ingestiondate=value)
        assert query == ""

    with pytest.raises(ValueError):
        SentinelAPI.format_query(date="NOW")

    with pytest.raises(ValueError):
        SentinelAPI.format_query(date=["NOW"])

    with pytest.raises(ValueError):
        SentinelAPI.format_query(ingestiondate=[])


@pytest.mark.vcr
@pytest.mark.scihub
def test_api_query_format_escape_spaces(api):
    query = SentinelAPI.format_query(ingestiondate=("NOW-1DAY", "NOW"))
    assert query == 'ingestiondate:["NOW-1DAY" TO "NOW"]'
    assert api.count(ingestiondate=("NOW-1DAY", "NOW")) > 0

    query = SentinelAPI.format_query(ingestiondate=("NOW-1DAY", None))
    assert query == 'ingestiondate:["NOW-1DAY" TO *]'
    assert api.count(ingestiondate=("NOW-1DAY", None)) > 0

    query = SentinelAPI.format_query(ingestiondate="[NOW-1DAY TO NOW]")
    assert query == "ingestiondate:[NOW-1DAY TO NOW]"

    query = SentinelAPI.format_query(ingestiondate=" [NOW-1DAY TO NOW] ")
    assert query == "ingestiondate:[NOW-1DAY TO NOW]"

    query = SentinelAPI.format_query(relativeorbitnumber="{101 TO 103}")
    assert query == "relativeorbitnumber:{101 TO 103}"
    assert api.count(relativeorbitnumber="{101 TO 103}") > 0

    query = SentinelAPI.format_query(filename="S3A_OL_2* ")
    assert query == "filename:S3A_OL_2*"

    query = SentinelAPI.format_query(filename="S3A_OL_2?2")
    assert query == "filename:S3A_OL_2?2"

    query = SentinelAPI.format_query(timeliness="Non Time Critical")
    assert query == r'timeliness:"Non Time Critical"'

    query = SentinelAPI.format_query(timeliness='"Non Time Critical"')
    assert query == r'timeliness:"Non Time Critical"'

    query = SentinelAPI.format_query(timeliness="Non\tTime\tCritical")
    assert query == r'timeliness:"Non Time Critical"'

    assert api.count(timeliness="Near Real Time") > 0

    # Allow for regex weirdness
    query = SentinelAPI.format_query(timeliness=".+ Critical")
    assert query == r'timeliness:".+ Critical"'
    assert api.count(timeliness=".+ Critical") > 0

    query = SentinelAPI.format_query(identifier="/S[123 ]A.*/")
    assert query == "identifier:/S[123 ]A.*/"
    kwargs = dict(beginposition=("20180101", "20180130"), area="POINT(14.20 59.96)")
    regex_count = api.count(identifier="/S[12 ]A.*/", **kwargs)
    c1 = api.count(identifier="S1A*", **kwargs)
    c2 = api.count(identifier="S2A*", **kwargs)
    assert min(regex_count, c1, c2) > 0
    assert regex_count == c1 + c2


@pytest.mark.vcr
@pytest.mark.scihub
def test_invalid_query(api):
    with pytest.raises(QuerySyntaxError):
        api.query(raw="xxx:yyy")


@pytest.mark.fast
def test_format_url(api):
    start_row = 0
    url = api._format_url(offset=start_row)
    assert (
        url
        == "https://apihub.copernicus.eu/apihub/search?format=json&rows={rows}&start={start}".format(
            rows=api.page_size, start=start_row
        )
    )
    limit = 50
    url = api._format_url(limit=limit, offset=start_row)
    assert (
        url
        == "https://apihub.copernicus.eu/apihub/search?format=json&rows={rows}&start={start}".format(
            rows=limit, start=start_row
        )
    )
    url = api._format_url(limit=api.page_size + 50, offset=start_row)
    assert (
        url
        == "https://apihub.copernicus.eu/apihub/search?format=json&rows={rows}&start={start}".format(
            rows=api.page_size, start=start_row
        )
    )
    url = api._format_url(order_by="beginposition desc", limit=api.page_size + 50, offset=10)
    assert (
        url == "https://apihub.copernicus.eu/apihub/search?format=json&rows={rows}&start={start}"
        "&orderby={orderby}".format(rows=api.page_size, start=10, orderby="beginposition desc")
    )


@pytest.mark.fast
def test_format_url_custom_api_url():
    api = SentinelAPI("user", "pw", api_url="https://scihub.copernicus.eu/dhus/")
    url = api._format_url()
    assert url.startswith("https://scihub.copernicus.eu/dhus/search")

    api = SentinelAPI("user", "pw", api_url="https://scihub.copernicus.eu/dhus")
    url = api._format_url()
    assert url.startswith("https://scihub.copernicus.eu/dhus/search")


@pytest.mark.fast
def test_format_order_by():
    res = _format_order_by("cloudcoverpercentage")
    assert res == "cloudcoverpercentage asc"

    res = _format_order_by("+cloudcoverpercentage,-beginposition")
    assert res == "cloudcoverpercentage asc,beginposition desc"

    res = _format_order_by(" +cloudcoverpercentage, -beginposition ")
    assert res == "cloudcoverpercentage asc,beginposition desc"

    with pytest.raises(ValueError):
        _format_order_by("+cloudcoverpercentage-beginposition")


@pytest.mark.vcr
@pytest.mark.scihub
def test_small_query(api, small_query):
    api.query(**small_query)
    assert api._last_query == (
        'beginPosition:["2015-01-01T00:00:00Z" TO "2015-01-02T00:00:00Z"] '
        'footprint:"Intersects(POLYGON((0 0,1 1,0 1,0 0)))"'
    )
    assert api._last_response.status_code == 200


@pytest.mark.vcr(decode_compressed_response=False)
@pytest.mark.scihub
def test_large_query(api, large_query):
    full_products = list(api.query(**large_query))
    assert api._last_query == (
        'beginPosition:["2015-12-01T00:00:00Z" TO "2015-12-31T00:00:00Z"] '
        'footprint:"Intersects(POLYGON((0 0,0 10,10 10,10 0,0 0)))"'
    )
    assert api._last_response.status_code == 200
    assert len(full_products) > api.page_size

    result = list(api.query(limit=150, **large_query))
    assert result == full_products[:150]

    result = list(api.query(limit=20, offset=90, **large_query))
    assert result == full_products[90:110]

    result = list(api.query(limit=20, offset=len(full_products) - 10, **large_query))
    assert result == full_products[-10:]


@pytest.mark.vcr
@pytest.mark.scihub
def test_count(api):
    count = api.count(None, ("20150101", "20151231"))
    assert count >= 10000


@pytest.mark.vcr
@pytest.mark.scihub
def test_too_long_query(api):
    # Test whether our limit calculation is reasonably correct and
    # that a relevant error message is provided

    def create_query(n):
        return '"' + " a_-.*:,?+~!" * n + '"'

    # Expect no error
    q = create_query(162)
    assert 0.99 < SentinelAPI.check_query_length(q) < 1.0
    count = api.count(raw=q)
    assert count == 0

    # Expect HTTP status 500 Internal Server Error
    q = create_query(163)
    assert 0.999 <= SentinelAPI.check_query_length(q) < 1.01
    with pytest.raises(QueryLengthError) as excinfo:
        api.count(raw=q)
    assert "x times the maximum allowed" in str(excinfo.value)


@pytest.mark.vcr
@pytest.mark.scihub
def test_date_arithmetic(api):
    products = api.query(
        "ENVELOPE(0, 1, 1, 0)", ("2016-12-01T00:00:00Z-1DAY", "2016-12-01T00:00:00Z+1DAY-1HOUR")
    )
    assert api._last_response.status_code == 200
    assert 0 < len(products) < 30


@pytest.mark.vcr
@pytest.mark.scihub
def test_quote_symbol_bug(api):
    # A test to check if plus symbol handling works correctly on the server side
    # It used to raise an error but has since been fixed
    # https://github.com/SentinelDataHub/DataHubSystem/issues/23

    q = "beginposition:[2017-05-30T00:00:00Z TO 2017-05-31T00:00:00Z+1DAY]"
    count = api.count(raw=q)
    assert count > 0


@pytest.mark.vcr
@pytest.mark.scihub
def test_s2_cloudcover(api, test_wkt):
    products = api.query(
        test_wkt, ("20181212", "20181228"), platformname="Sentinel-2", cloudcoverpercentage=(0, 10)
    )

    product_ids = list(products)
    assert product_ids == [
        "bf652bc4-299c-4c39-9238-ee5a3fdc0d3e",
        "b508a8cd-c7d6-4a4d-9286-6d9463926554",
        "c9b0a744-c0e5-41f9-af6c-f0af83681e58",
        "2e69293b-591f-41d4-99e5-89ec087ae487",
        "dcd0849f-f43a-46f4-9267-da8069b74dd8",
    ]

    # For order-by test
    vals = [x["cloudcoverpercentage"] for x in products.values()]
    assert sorted(vals) != vals
    assert all(0 <= x <= 10 for x in vals)


@pytest.mark.vcr
@pytest.mark.scihub
def test_order_by(api, test_wkt):
    kwargs = dict(
        area=test_wkt,
        date=("20151219", "20160519"),
        platformname="Sentinel-2",
        cloudcoverpercentage=(0, 10),
        order_by="cloudcoverpercentage, -beginposition",
    )
    # Check that order_by works correctly also in cases where pagination is required
    expected_count = api.count(**kwargs)
    assert 100 < expected_count < 250
    products = api.query(**kwargs)
    assert len(products) == expected_count
    vals = [x["cloudcoverpercentage"] for x in products.values()]
    assert sorted(vals) == vals
    assert all(0 <= x <= 10 for x in vals)


@pytest.mark.vcr
@pytest.mark.scihub
def test_area_relation(api):
    params = dict(
        area="POLYGON((10.83 53.04,11.64 53.04,11.64 52.65,10.83 52.65,10.83 53.04))",
        date=("20151219", "20151226"),
    )
    result = api.query(**params)
    n_intersects = len(result)
    assert n_intersects > 10

    result = api.query(area_relation="contains", **params)
    n_contains = len(result)
    assert 0 < n_contains < n_intersects
    result = api.query(area_relation="IsWithin", **params)
    n_iswithin = len(result)
    assert n_iswithin == 0

    # Check that unsupported relations raise an error
    with pytest.raises(ValueError):
        api.query(area_relation="disjoint", **params)


@pytest.mark.vcr
@pytest.mark.scihub
def test_query_by_names(api, smallest_online_products):
    names = [product["title"] for product in smallest_online_products]
    expected = {product["title"]: {product["id"]} for product in smallest_online_products}

    result = api._query_names(names)
    assert list(result) == names
    for name in names:
        assert set(result[name]) == expected[name]

    result2 = api._query_names(names * 3)
    assert result == result2


@pytest.mark.fast
def test_empty_query(api):
    with pytest.raises(ValueError):
        api.query()



import re
from os.path import join

from vcr.serializers import yamlserializer


class BinaryContentSerializer:
    """
    Serializer that stores any binary content in queries on disk as separate files
    """

    def __init__(self, directory=".", base_serializer=yamlserializer):
        self.directory = directory
        self.base_serializer = base_serializer

    def deserialize(self, cassette_string):
        cassette_dict = self.base_serializer.deserialize(cassette_string)
        for interaction in cassette_dict["interactions"]:
            if interaction["request"]["method"] == "HEAD":
                continue
            response = interaction["response"]
            headers = {k.lower(): v for k, v in response["headers"].items()}
            if "content-range" in headers and "content-disposition" in headers:
                rg, size, filename = self._parse_headers(headers)
                with open(join(self.directory, "data", filename), "rb") as f:
                    f.seek(rg[0])
                    content = f.read(rg[1] - rg[0] + 1)
                response["body"]["string"] = content
        return cassette_dict

    def serialize(self, cassette_dict):
        for interaction in cassette_dict["interactions"]:
            if interaction["request"]["method"] == "HEAD":
                continue
            response = interaction["response"]
            headers = {k.lower(): v for k, v in response["headers"].items()}
            if "content-range" in headers and "content-disposition" in headers:
                rg, size, filename = self._parse_headers(headers)
                content = response["body"]["string"]
                if hasattr(content, "encode"):
                    content = content.encode("utf-8")
                if rg[0] == 0 and rg[1] + 1 == size:
                    with open(join(self.directory, "data", filename), "wb") as f:
                        f.write(content)
                del response["body"]["string"]
        return self.base_serializer.serialize(cassette_dict)

    @staticmethod
    def _parse_headers(headers):
        range_hdr = headers["content-range"][0]
        m = re.match(r"^bytes (\d+)-(\d+)/(\d+)", range_hdr)
        rg = (int(m.group(1)), int(m.group(2)))
        size = int(m.group(3))
        filename = headers["content-disposition"][0].split('"')[1]
        return rg, size, filename



import fnmatch
from pathlib import Path
from xml.etree import ElementTree as etree

import sentinelsat
from sentinelsat.exceptions import LTATriggered, SentinelAPIError
from sentinelsat.sentinel import _check_scihub_response


class SentinelProductsAPI(sentinelsat.SentinelAPI):
    """Class to connect to Copernicus Open Access Hub, search and download imagery.

    The products node interface allows to filter and download individual product
    files by means of a (optional) *nodefilter* callable function.
    For each file in the product (only excluding the manifest) the *nodefilter* function
    is called to decide if the corresponding file must be downloaded or not.

    The *nodefilter* function has the following signature::

      def nodefilter(node_info: dict) -> bool:
          ...

    The *node_info* parameter is a dictionary containing information like

    * the file *path* within the product (e.g. "./preview/map-overlay.kml")
    * the file size in bytes (int)
    * the file md5

    It the *nodefilter* function returns True the corresponding file is downloaded,
    otherwise the file is not downloaded.


    Parameters
    ----------
    user : string
        username for DataHub
        set to None to use ~/.netrc
    password : string
        password for DataHub
        set to None to use ~/.netrc
    api_url : string, optional
        URL of the DataHub
        defaults to 'https://apihub.copernicus.eu/apihub'
    show_progressbars : bool
        Whether progressbars should be shown or not, e.g. during download. Defaults to True.
    timeout : float or tuple, optional
        How long to wait for DataHub response (in seconds).
        Tuple (connect, read) allowed.

    Attributes
    ----------
    session : requests.Session
        Session to connect to DataHub
    api_url : str
        URL to the DataHub
    page_size : int
        Number of results per query page.
        Current value: 100 (maximum allowed on ApiHub)
    timeout : float or tuple
        How long to wait for DataHub response (in seconds).


    .. versionadded:: 0.15
    """

    def _path_to_url(self, product_info, path, urltype=None):
        id = product_info["id"]
        title = product_info["title"]
        path = "/".join(["Nodes('{}')".format(item) for item in path.split("/")])
        if urltype == "value":
            urltype = "/$value"
        elif urltype == "json":
            urltype = "?$format=json"
        elif urltype == "full":
            urltype = "?$format=json&$expand=Attributes"
        elif urltype is None:
            urltype = ""
        # else: pass urltype as is
        return self._get_odata_url(id, f"/Nodes('{title}.SAFE')/{path}{urltype}")

    def _get_manifest(self, product_info, path=None):
        path = Path(path) if path else None
        url = self._path_to_url(product_info, "manifest.safe", "value")
        node_info = product_info.copy()
        node_info["url"] = url
        node_info["node_path"] = "./manifest.safe"
        del node_info["md5"]

        if path and path.exists():
            self.logger.info("manifest file already available (%r), skip download", path)
            data = path.read_bytes()
            node_info["size"] = len(data)
            return node_info, data

        url = self._path_to_url(product_info, "manifest.safe", "json")
        response = self.session.get(url, auth=self.session.auth)
        _check_scihub_response(response)
        info = response.json()["d"]

        node_info["size"] = int(info["ContentLength"])

        response = self.session.get(node_info["url"], auth=self.session.auth)
        _check_scihub_response(response, test_json=False)
        data = response.content
        if len(data) != node_info["size"]:
            raise SentinelAPIError("File corrupt: data length do not match")

        if path:
            path.parent.mkdir(parents=True, exist_ok=True)
            path.write_bytes(data)

        return node_info, data

    def _dataobj_to_node_info(self, dataobj_info, product_info):
        path = dataobj_info["href"]
        if path.startswith("./"):
            path = path[2:]

        node_info = product_info.copy()
        node_info["url"] = self._path_to_url(product_info, path, "value")
        node_info["size"] = dataobj_info["size"]
        node_info["md5"] = dataobj_info["md5"]
        node_info["node_path"] = dataobj_info["href"]
        # node_info["parent"] = product_info

        return node_info

    def _filter_nodes(self, manifest, product_info, nodefilter=None):
        nodes = {}
        xmldoc = etree.parse(manifest)
        data_obj_section_elem = xmldoc.find("dataObjectSection")
        for elem in data_obj_section_elem.iterfind("dataObject"):
            dataobj_info = _xml_to_dataobj_info(elem)
            node_info = self._dataobj_to_node_info(dataobj_info, product_info)
            if nodefilter is not None and not nodefilter(node_info):
                continue
            node_path = node_info["node_path"]
            nodes[node_path] = node_info
        return nodes

    def download(self, id, directory_path=".", checksum=True, nodefilter=None, **kwargs):
        """Download a product.

        Uses the filename on the server for the downloaded files, e.g.
        "S1A_EW_GRDH_1SDH_20141003T003840_20141003T003920_002658_002F54_4DD1.SAFE/manifest.safe".

        Incomplete downloads are continued and complete files are skipped.

        Parameters
        ----------
        id : string
            UUID of the product, e.g. 'a8dd0cfd-613e-45ce-868c-d79177b916ed'
        directory_path : string, optional
            Where the file will be downloaded
        checksum : bool, optional
            If True, verify the downloaded file's integrity by checking its MD5 checksum.
            Throws InvalidChecksumError if the checksum does not match.
            Defaults to True.
        nodefilter : callable, optional
            The *nodefilter* callable used to select which file of each product have to
            be downloaded.
            If *nodefilter* is None then no file filtering is performed and the class
            behaves exactly as :class:`sentinelsat.sentinel.SentinelAPI`.


        Returns
        -------
        product_info : dict
            Dictionary containing the product's info from get_product_info() as well as
            the path on disk.

        Raises
        ------
        InvalidChecksumError
            If the MD5 checksum does not match the checksum on the server.
        """
        if nodefilter is None:
            return sentinelsat.SentinelAPI.download(self, id, directory_path, checksum, **kwargs)

        product_info = self.get_product_odata(id)
        product_path = Path(directory_path) / (product_info["title"] + ".SAFE")
        product_info["node_path"] = "./" + product_info["title"] + ".SAFE"
        manifest_path = product_path / "manifest.safe"
        if not manifest_path.exists() and self.trigger_offline_retrieval(id):
            raise LTATriggered(id)

        manifest_info, _ = self._get_manifest(product_info, manifest_path)
        product_info["nodes"] = {
            manifest_info["node_path"]: manifest_info,
        }

        node_infos = self._filter_nodes(manifest_path, product_info, nodefilter)
        product_info["nodes"].update(node_infos)

        for node_info in node_infos.values():
            node_path = node_info["node_path"]
            path = (product_path / node_path).resolve()
            node_info["path"] = path
            node_info["downloaded_bytes"] = 0

            self.logger.info("Downloading %s node to %s", id, path)
            self.logger.debug("Node URL for %s: %s", id, node_info["url"])

            if path.exists():
                # We assume that the product node has been downloaded and is complete
                continue

            self._download_outer(node_info, path, checksum)

        return product_info


def _xml_to_dataobj_info(element):
    assert etree.iselement(element)
    assert element.tag == "dataObject"
    data = dict(
        id=element.attrib["ID"],
        rep_id=element.attrib["repID"],
    )
    elem = element.find("byteStream")
    # data["mime_type"] = elem.attrib['mimeType']
    data["size"] = int(elem.attrib["size"])
    elem = element.find("byteStream/fileLocation")
    data["href"] = elem.attrib["href"]
    # data['locator_type'] = elem.attrib["locatorType"]
    # assert data['locator_type'] == "URL"

    elem = element.find("byteStream/checksum")
    assert elem.attrib["checksumName"].upper() == "MD5"
    data["md5"] = elem.text

    return data


def make_size_filter(max_size):
    """Generate a nodefilter function to download only files below the specified maximum size.

    .. versionadded:: 0.15
    """

    def node_filter(node_info):
        return node_info["size"] <= max_size

    return node_filter


def make_path_filter(pattern, exclude=False):
    """Generate a nodefilter function to download only files matching the specified pattern.

    Parameters
    ----------
    pattern : str
        glob patter for files selection
    exclude : bool, optional
        if set to True then files matching the specified pattern are excluded. Default False.

    .. versionadded:: 0.15
    """

    def node_filter(node_info):
        match = fnmatch.fnmatch(node_info["node_path"].lower(), pattern)
        return not match if exclude else match

    return node_filter


def all_nodes_filter(node_info):
    """Node filter function to download all files.

    This function can be used to download Sentinel product as a directory
    instead of downloading a single zip archive.

    .. versionadded:: 0.15
    """
    return True
